import os
from kevin_toolbox.data_flow.core.reader import File_Iterative_Reader
from kevin_toolbox.data_flow.file.kevin_notation.converter import Converter, CONVERTER_FOR_READER


class Kevin_Notation_Reader:
    """
        遵守 kevin_notation 格式的数据文本读取器（格式要求参见本模块下的 readme）
            支持分批次读取文件内容
    """

    def __init__(self, **kwargs):
        """
            设定关键参数

            必要参数：
                file_path:          <string> 文件路径
            读取相关参数：
                chunk_size:         <integer> 每次读取多少行数据
                beg：                <integer> 开始读取的位置
                                                默认为0
                converter:          <instance of kevin_toolbox.Converter> converter is a dictionary-like data structure
                                                consisting of <string>:<func> pairs，
                                                用于根据指定数据类型选取适当的函数来处理输入数据。
        """

        # 默认参数
        paras = {
            # 必要参数
            "file_path": None,
            # 读取相关参数
            "chunk_size": 100,
            "beg": 0,
            "converter": CONVERTER_FOR_READER,
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        assert isinstance(paras["file_path"], (str,)) and os.path.isfile(paras["file_path"]), \
            f'file not exists :{paras["file_path"]}'
        #
        assert isinstance(paras["chunk_size"], (int,)) and (paras["chunk_size"] > 0 or paras["chunk_size"] == -1)
        assert isinstance(paras["beg"], (int,)) and paras["beg"] >= 0
        assert isinstance(paras["converter"], (Converter, dict,))

        self.paras = paras

        # 读取开头
        self.reader = File_Iterative_Reader(file_path=self.paras["file_path"],
                                            pre_jump_size=self.paras["beg"],
                                            filter_=lambda x: x != "\n" and not x.startswith("//"),  # 去除注释
                                            map_func=lambda x: x.rsplit("\n", 1)[0].split("//", 1)[0],
                                            drop=False)
        # kevin_notation
        offset = self.read_head(reader=self.reader, head="# --kevin_notation--")
        # metadata
        offset += self.read_head(reader=self.reader, head="# --metadata--")
        #
        self.metadata, count = self.read_metadata(reader=self.reader, converter=self.paras["converter"])
        offset += count

        del self.reader

        # 读取内容
        self.reader = File_Iterative_Reader(file_path=self.paras["file_path"],
                                            pre_jump_size=self.paras["beg"] + offset,
                                            filter_=lambda x: x != "\n" and not x.startswith("//"),  # 去除注释
                                            map_func=lambda x: x.rsplit("\n", 1)[0].split("//", 1)[0],
                                            drop=False)

        # contents
        self.read_head(reader=self.reader, head="# --contents--")

    # ------------------------------------ read head ------------------------------------ #

    @staticmethod
    def read_head(reader, head, try_times=10):
        reader.paras.update(dict(
            chunk_size=1,
        ))
        #
        for count in range(1, try_times + 1):
            line = next(reader)[0]
            if line.startswith("#"):
                assert line.startswith(head), \
                    f"format of {head} is wrong!"
                return count
        raise Exception(f"can't find {head} after {try_times} attempts")

    # ------------------------------------ read metadata ------------------------------------ #

    @staticmethod
    def read_metadata(reader, converter):
        reader.paras.update(dict(
            chunk_size=2,
        ))
        key, sep = next(reader)
        assert key.startswith("# sep"), \
            f"format of metadata is wrong!"  # 开头第一个参数必须是 sep
        res = dict(sep=sep)
        #
        count = 2
        for chunk in reader:
            if len(chunk) == 2:
                key, value = chunk
            else:
                key, value = chunk[0], None
            assert key.startswith("# "), f"{key}:{value}"
            if key.startswith("# --contents--"):
                break
            temp_ls = [i.strip() for i in key.split(' ', 2)]
            assert len(temp_ls) >= 2
            #
            count += 2
            #
            key = temp_ls[1]
            #
            paras_dict = dict()
            if len(temp_ls) > 2:
                paras = temp_ls[2]
                assert paras.startswith("(") and paras.endswith(")")
                paras = paras[1:-1]
                paras_ls = [i.split("=", 1) for i in paras.split(",", -1)]
                paras_dict.update({k: v for (k, v) in paras_ls})
            sep = paras_dict.get("sep", res["sep"])
            #
            if sep in value:
                value = value.split(sep, -1)
            #
            res[key] = value

        # 最后检验内容
        # 与 column 数量相关的字段是否相互一致
        num_ls = [len(res[key]) for key in ["column_name", "column_type"] if key in res]
        num_ls.extend([int(res[key]) for key in ["column_num"] if key in res])
        assert len(set(num_ls)) <= 1, f"{num_ls}"
        #
        if len(num_ls) > 0:
            res["column_num"] = num_ls[0]
        #
        if "column_type" in res:
            unknown_types = set(res["column_type"]).difference(set(converter.keys()))
            if len(unknown_types) > 0:
                print(f"Warning: There are unknown types {unknown_types} in column_type. "
                      f"Currently supported types are {set(converter.keys())}")

        return res, count

    # ------------------------------------ read contents ------------------------------------ #

    @staticmethod
    def read_contents(reader, chunk_size, metadata, converter):
        """
            读取 contents，并根据 metadata 对其进行解释
        """
        reader.paras.update(dict(
            chunk_size=chunk_size,
        ))
        #
        res = None
        chunk = next(reader)
        # 解释
        for line in chunk:
            value_ls = line.split(metadata["sep"])
            if "column_type" in metadata:
                assert len(value_ls) == len(metadata["column_type"])
                type_ls = metadata["column_type"]
            else:
                # use default converter
                type_ls = ["default"] * len(value_ls)
            value_ls = [converter[type_](value) for type_, value in zip(type_ls, value_ls)]
            if res is None:
                res = [[] for _ in range(len(value_ls))]
            for i, value in enumerate(value_ls):
                res[i].append(value)
        # 整理
        if res is not None:
            column_name = metadata.get("column_name", list(range(len(res))))
            assert len(column_name) == len(res)
            res = {k: res[i] for i, k in enumerate(column_name)}
        return res, len(chunk)

    # ------------------------------------ magic func ------------------------------------ #

    def __next__(self):
        contents, len_ = self.read_contents(reader=self.reader, chunk_size=self.paras["chunk_size"],
                                            metadata=self.metadata, converter=self.paras["converter"])
        if len_ == 0:
            raise StopIteration
        else:
            return contents

    def __iter__(self):
        return self

    # with 上下文管理器
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self

    def __del__(self):
        try:
            del self.paras, self.metadata, self.reader
        except Exception as e:
            print(e)



import numpy as np


def softmax(v, axis=-1):
    # 为了数值稳定，减去最大值
    v = v - np.max(v)
    return np.exp(v) / np.sum(np.exp(v), axis=axis, keepdims=True)


if __name__ == '__main__':
    print(softmax(np.asarray([0, 0.1]) * 10))



def parse_version_string_to_array(string, **kwargs):
    """
        将版本的字符串转换为数组的形式

        参数：
            string:                 string of version
            sep：                    分隔符
        工作流程：
            string "1.2.3"  == sep '.' ==> array [1, 2, 3]
            注意：对于分割后包含非整数的 string，返回 -1，例如：
            string "1.2_3"  == sep '_' ==> array [-1, 3]
        返回：
            array:                 array of version
    """
    sep = kwargs.get("sep", '.')
    assert isinstance(string, (str,))

    array = string.split(sep, -1)
    array = [int(v) if v.isdigit() else -1 for v in array]
    return array


if __name__ == '__main__':
    print(parse_version_string_to_array("1.2.3"))
    print(parse_version_string_to_array("1.2_3", sep="_"))



from kevin_toolbox.data_flow.file import kevin_notation


def read(file_path):
    """
        读取整个文件的快捷接口
    """
    with kevin_notation.Reader(file_path=file_path, chunk_size=-1) as reader:
        # metadata
        metadata = reader.metadata
        # content
        try:
            content = next(reader)
        except:
            content = {key: [] for key in metadata["column_name"]}
    return metadata, content



import numpy as np
import torch
from kevin_toolbox.patches.for_torch.compatible import tile as torch_tile
from kevin_toolbox.data_flow.core.reader import Unified_Reader_Base, UReader

SUPPORT_TO_GENERATE = {"scores", "labels", "samples"}


class Face_Verification_DataSet_Factory:
    """
        用于生成人脸识别 1:1 验证任务的数据集

        背景介绍：
            人脸识别任务中的 1:1 人脸验证，其实就可以看做一个二分类任务。
            我们将两个人脸图片之间的一次比较（注意这里的“比较”是名词，我们也可以将它看做由两个人脸图片组成的二元对），看做一个样本 sample。
            - 同一个人的比较，看做 positive sample
            - 不同人之间的比较，看做 negative sample

        人脸特征相关变量（按照 feature_id 进行排序，一一对应）：
            features:        list of feature
                                    人脸特征库
                                        shape [feature_nums, feature_dims]
            clusters:     list of feature' cluster id
                                    记录各个feature所属的cluster（亦即人）
                                        shape [feature_nums, 1]

        数据集相关变量（按照 sample_id 进行排序，一一对应）：
            samples:                list of feature_id pairs
                                    每个样本由一对 feature_id 构成，表示这两个feature之间的一次比较。
                                        shape [sample_nums, 2]
            scores:                 list of floats indicating similarities between feature pairs
                                    两个feature之间的相似度
                                        shape [sample_nums, 1]
            labels:                 list of labels describing whether it is a positive or negative sample
                                    描述是正样本还是负样本（亦即是不是同一个cluster/人）
                                        shape [sample_nums, 1]

        工作流程：
            根据 feature_id 来生成 samples
            根据 features 来生成 scores
            根据 clusters 来生成 labels

        用法：
            1. __init__() 在初始化时，输入人脸特征相关变量
                factory = Face_Verification_DataSet_Factory( features=xxx, clusters=xxx )
                    要求输入的 features 和 clusters 需要使用 Unified_Reader_Base 的实现来包裹
            2. 通过 generate_xxx 系列函数来生成数据集
                # 方式 1
                #   generate_by_block
                #   先根据以 feature_id 为行列的相互矩阵来构建 samples 样本对，
                #   在该矩阵中所有的 feature 两两结对，每个格点对应于一个 sample，
                #   然后可以通过指定矩阵中的一个子矩阵来决定要生成哪一部分的数据集。
                # 有两种指定方式：
                #   连续行列
                dataset = factory.generate_by_block(i_0, i_1, j_0, j_1)
                #   不连续行列
                dataset = factory.generate_by_block(i_ls, None, j_ls, None)
                #   也可以混合指定
                dataset = factory.generate_by_block(i_ls, None, j_0, j_1)

                # 方式 2
                #   generate_by_samples
                #   根据输入的 samples 来生成数据集
                dataset = factory.generate_by_samples(samples)

                # 返回的数据集是一个字典
                # dataset := {"scores": xxx, "labels": xxx, "samples": xxx}
                # 其中具体包含哪些字段，可以通过添加 need_to_generate 参数来指定
                # 目前支持的字段有： {"scores", "labels", "samples"}
                dataset = factory.generate_by_samples(samples, need_to_generate={"scores", "samples"})
    """

    def __init__(self, *args, **kwargs):
        """
            绑定人脸特征相关变量
            必要参数:
                features:               list of feature 人脸特征库
                                            shape [feature_nums, feature_dims]
            可选参数：
                feature_id:             list of feature' id
                                            shape [feature_nums, 1]
                                            不指定时，默认从0开始依次递增
                                            要求不能有重复
                clusters:               list of feature' cluster id 记录各个feature所属的cluster（亦即人）
                                            shape [feature_nums, 1]
                                            不指定时，默认将各个 feature 看作从属于各自独立的 cluster，亦即使用 feature_id 来填充
                注意：以上输入的数据都需要被 Unified_Reader_Base 包裹
                    而 Unified_Reader_Base 除了可以支持从内存变量中读取数据外，还支持从持久化的磁盘文件中读取数据
            其他参数：
                feature_ids_is_sequential:       boolean，feature_ids 是否以1为间距递增的
                                            默认 False
                                            具体作用参考 generate_by_samples() 中的介绍
        """

        # 默认参数
        paras = {
            # 人脸特征相关变量
            "features": None,
            "feature_ids": None,
            "clusters": None,
            # 补充
            "feature_ids_is_sequential": False,
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        # features [feature_nums, feature_dims]
        assert isinstance(paras["features"], (Unified_Reader_Base,))
        # feature_id [feature_nums, 1]
        if paras["feature_ids"] is not None:
            assert isinstance(paras["feature_ids"], (Unified_Reader_Base,))
        else:
            # 自动生成
            paras["feature_ids"] = UReader(var=np.arange(len(paras["features"])).reshape((-1, 1)))
            # 默认生成的 feature_ids 是连续的
            paras["feature_ids_is_sequential"] = True
        # clusters [feature_nums, 1]
        if paras["clusters"] is not None:
            assert isinstance(paras["clusters"], (Unified_Reader_Base,)) and len(
                paras["clusters"]) == len(paras["features"])
        else:
            # 自动生成
            paras["clusters"] = paras["feature_ids"]

        self.paras = paras

        # 计算设备（尽量使用gpu来加速计算）
        self.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

    @staticmethod
    def __check_paras_of_block(i_0, i_1, j_0, j_1, **kwargs):
        """
            检查 generate_by_block() 的输入参数
        """
        # (i_0, i_1, j_0, j_1)
        i_len = i_1 - i_0 if isinstance(i_0, (int,)) else len(i_0)
        j_len = j_1 - j_0 if isinstance(j_0, (int,)) else len(j_0)
        assert i_len > 0 and j_len > 0

        # pick_triangle
        pick_triangle = kwargs.get("pick_triangle", False)

        # include_diagonal
        include_diagonal = kwargs.get("include_diagonal", True)

        # need_to_generate
        need_to_generate = kwargs.get("need_to_generate", None)
        if need_to_generate is not None:
            assert isinstance(need_to_generate, (set,)) and need_to_generate.issubset(SUPPORT_TO_GENERATE), \
                ValueError(f"need_to_generate should be a subset of {SUPPORT_TO_GENERATE}")
        else:
            need_to_generate = SUPPORT_TO_GENERATE

        return i_len, j_len, pick_triangle, include_diagonal, need_to_generate

    def generate_by_block(self, i_0, i_1, j_0, j_1, **kwargs):
        """
            根据 feature_id 的相互矩阵来构建数据集
                先根据以 feature_id 为行列的相互矩阵来构建 samples 样本对，在该矩阵中所有的 feature 两两结对，
                每个格点对应于一个 sample，然后可以通过指定矩阵中的一个子矩阵来决定要生成哪一部分的数据集。

            范围参数：
                有两种指定方式：
                连续行列
                    generate_by_block(i_0, i_1, j_0, j_1)
                不连续行列
                    generate_by_block(i_ls, None, j_ls, None)
                也可以混合指定
                    generate_by_block(i_ls, None, j_0, j_1)

            其他参数：
                pick_triangle:      只取block的上三角
                include_diagonal:   当取上三角时，是否包含对角线
                                        默认为 True 包含
                                        仅当 pick_triangle=True 时该参数起效
                need_to_generate:   指定数据集中需要生成的字段
                                        目前支持的字段有： {"scores", "labels", "samples"}
                                        
            补充：
                当参数中额外指定有 scores, labels 或者 samples 时，将以参数中指定的值为准，不对该项进行计算
        """
        i_len, j_len, pick_triangle, include_diagonal, need_to_generate = self.__check_paras_of_block(i_0, i_1,
                                                                                                      j_0, j_1,
                                                                                                      **kwargs)
        res = dict()
        for key in need_to_generate:
            if key in kwargs:
                # 以参数指定的为准
                # 不需要进行计算
                res[key] = kwargs[key]
                continue

            # 计算 scores
            if key == "scores":
                # feature_outer [i_len, feature_dims]  and  [j_len, feature_dims]
                feature_outer_i = torch.tensor(self.paras["features"].read(i_0, i_1), device=self.device,
                                               dtype=torch.float32)
                feature_outer_j = torch.tensor(self.paras["features"].read(j_0, j_1), device=self.device,
                                               dtype=torch.float32)
                # scores [i_len, j_len, 1]
                res["scores"] = feature_outer_i.matmul(feature_outer_j.t()).unsqueeze(-1)
            # 计算 labels
            elif key == "labels":
                # cluster_outer [i_len, 1]  and  [j_len, 1]
                cluster_outer_i = torch.tensor(self.paras["clusters"].read(i_0, i_1), device=self.device,
                                               dtype=torch.float32)
                cluster_outer_j = torch.tensor(self.paras["clusters"].read(j_0, j_1), device=self.device,
                                               dtype=torch.float32)
                # labels [i_len, j_len, 1]
                res["labels"] = (cluster_outer_i.reshape(-1, 1) == cluster_outer_j.reshape(1, -1)).unsqueeze(-1)
            # 计算 samples
            elif key == "samples":
                # fid_outer [i_len, 1]  and  [j_len, 1]
                fid_outer_i = torch.tensor(self.paras["feature_ids"].read(i_0, i_1), device=self.device,
                                           dtype=torch.float32)
                fid_outer_j = torch.tensor(self.paras["feature_ids"].read(j_0, j_1), device=self.device,
                                           dtype=torch.float32)
                # samples [i_len, j_len, 2]
                res["samples"] = torch.stack(
                    (torch_tile(fid_outer_i, (1, j_len)), torch_tile(fid_outer_j.t(), (i_len, 1))), dim=2)

        # reshape 并转换为 np.array
        for key in need_to_generate:
            if pick_triangle:
                # 取上三角部分
                offset = 0 if include_diagonal else 1  # 是否包含对角线
                indices = tuple(torch.triu_indices(row=i_len, col=j_len, offset=offset))
                res[key] = res[key][indices].cpu().numpy()
            else:
                shape = [i_len * j_len] + list(res[key].shape[2:])
                res[key] = res[key].reshape(shape).cpu().numpy()

        return res

    def cal_size_of_block(self, i_0, i_1, j_0, j_1, **kwargs):
        """
            计算 generate_by_block() 将会产生的数据集的大小
            参数：
                与 generate_by_block() 完全相同
        """
        i_len, j_len, pick_triangle, include_diagonal, _ = self.__check_paras_of_block(i_0, i_1, j_0, j_1, **kwargs)
        if pick_triangle:
            offset = 0 if include_diagonal else 1
            size = (j_len - offset) * (j_len + 1 - offset)
            if i_len < j_len:
                size -= (j_len - i_len - offset) * (j_len - i_len + 1 - offset)
            size /= 2
        else:
            size = i_len * j_len
        return int(size)

    def __check_paras_of_samples(self, samples, **kwargs):
        """
            检查 generate_by_samples() 的输入参数
        """
        # samples
        assert samples.ndim == 2 and samples.shape[1] == 2, \
            Exception(f"Error: shape {samples.shape} of samples does not satisfy [sample_nums, 2]!")

        # need_to_generate
        need_to_generate = kwargs.get("need_to_generate", None)
        if need_to_generate is not None:
            assert isinstance(need_to_generate, (set,)) and need_to_generate.issubset(SUPPORT_TO_GENERATE), \
                ValueError(f"need_to_generate should be a subset of {SUPPORT_TO_GENERATE}")
        else:
            need_to_generate = SUPPORT_TO_GENERATE

        # feature_ids_is_sequential
        feature_id_is_sequential = kwargs.get("feature_id_is_sequential", self.paras["feature_ids_is_sequential"])
        return samples, need_to_generate, feature_id_is_sequential

    def generate_by_samples(self, samples, **kwargs):
        """
            根据输入的 samples 来生成数据集
                通过 samples 中的 feature_id 到 feature_ids 中找到对应的index，然后根据该index到 features 和 clusters 中取对应的值。

            范围参数：
                samples:            list of feature_id pairs
                                        np.array with dtype=np.int
                                        shape [sample_nums, 2]
            其他参数：
                need_to_generate:   参见 generate_by_block() 中的介绍。
                feature_ids_is_sequential:       boolean，feature_ids 是否以1为间距递增的
                                        默认 False
                                        当设定为 True 时，亦即 feature_ids 以1为间距递增的，
                                        那么我们在找到 features 中各个元素的 index 与它的 feature_id 之间的偏移关系后，
                                        对于任意的 feature_id，就可以直接根据找到的关系式 index = feature_id + offset，
                                        检索到对应的 feature = features[index]，提高检索效率。
                                        
            补充：
                当参数中额外指定有 scores, labels 或者 samples 时，将以参数中指定的值为准，不对该项进行计算
        """
        samples, need_to_generate, feature_id_is_sequential = self.__check_paras_of_samples(samples, **kwargs)

        # 获取 index
        if feature_id_is_sequential:
            offset = self.paras["feature_ids"].read(0)[0]
            samples_ = samples - offset
            index_ls_i, index_ls_j = list(samples_[:, 0]), list(samples_[:, 1])
        else:
            map_ = dict().fromkeys(samples.flat)  # { feature_id: index, ... }
            for key in map_.keys():
                map_[key] = self.paras["feature_ids"].find(key)
            #
            index_ls_i, index_ls_j = [], []
            for i, j in samples:
                index_ls_i.append(map_[i])
                index_ls_j.append(map_[j])

        res = dict()
        for key in need_to_generate:
            if key in kwargs:
                # 以参数指定的为准
                # 不需要进行计算
                res[key] = kwargs[key]
                continue

            # 计算 scores
            if key == "scores":
                # features [nums, feature_dims]
                features_i = torch.tensor(self.paras["features"].read(index_ls_i), device=self.device,
                                          dtype=torch.float32)
                features_j = torch.tensor(self.paras["features"].read(index_ls_j), device=self.device,
                                          dtype=torch.float32)
                # scores [nums, 1]
                scores = torch.sum(features_i * features_j, dim=1, keepdim=True)
                res["scores"] = scores.cpu().numpy()
            # 计算 labels
            elif key == "labels":
                # clusters [nums, 1]
                clusters_i = torch.tensor(self.paras["clusters"].read(index_ls_i), device=self.device,
                                          dtype=torch.float32)
                clusters_j = torch.tensor(self.paras["clusters"].read(index_ls_j), device=self.device,
                                          dtype=torch.float32)
                # labels [nums, 1]
                labels = clusters_i == clusters_j
                res["labels"] = labels.cpu().numpy()
            # 计算 samples
            elif key == "samples":
                # samples [nums, 2]
                res["samples"] = samples

        return res



import torch
from kevin_toolbox.env_info import version

"""
    在 1.8 及其以下版本的 pytorch 中，没有 torch.linalg.norm()
        改用 torch.norm() 来实现，
        但是需要注意两者的输出并非完全一致，会有一定的偏差，但非常微小（在1e-10的量级）
"""

if version.compare(torch.__version__, "<=", "1.8", mode="short"):
    _arg_name_ls = [None, "ord", "dim", "keepdim", "out", "dtype"]


    def norm(*args, **kwargs):
        if len(args) > 0:
            X = args[0]
            for name, arg in zip(_arg_name_ls[1:len(args)], args[1:]):
                kwargs[name] = arg
            args = [X]
        kwargs["p"] = kwargs.pop("ord")
        return torch.norm(*args, **kwargs)
else:
    norm = torch.linalg.norm



import matplotlib.axes._axes as axes
import numpy as np


def add_trajectory_2d(ax, points, draw_key_points=True, scale=None):
    """
        参数：
            ax:
            points:                 <np.array> 坐标点 shape [batch, 2]
            draw_key_points:        <boolean> 是否绘制关键点
            scale:                  <float> 放大系数
                                        不设定时，默认根据 points 推断得到的最小分辨率来设置。
    """
    assert isinstance(ax, (axes.Axes,))
    points = np.asarray(points)
    assert len(points) >= 1 and points.ndim == 2 and points.shape[-1] == 2, \
        f'expected shape of points is [batch, 3], but got a {points.shape}'

    if scale is None:
        if len(points) > 1:
            temp = points.copy()
            temp[1:] -= temp[:-1]
            temp[0] -= points[-1]
            scale = np.min(np.sum(temp ** 2, axis=1)) ** 0.5
        else:
            scale = 1

    # 箭头
    for beg, end in zip(points[:-1], points[1:]):
        # ax.arrow(x=beg[0], y=beg[1], dx=end[0] - beg[0], dy=end[1] - beg[1],
        #          width=scale * 0.05, head_width=scale * 0.2, overhang=scale * 0.2,
        #          color="red", joinstyle="round",
        #          length_includes_head=True)
        # ax.quiver(*beg.tolist(), *(end - beg).tolist(), scale=5, width=0.01, headwidth=4, color="blue", alpha=0.5)
        ax.annotate("", xy=end, xytext=beg,
                    arrowprops=dict(width=scale * 0.5, headwidth=scale * 6, color="blue", alpha=0.5))

    # 关键点
    if draw_key_points:
        ax.scatter(x=points[:, 0], y=points[:, 1],
                   marker='o', s=scale * 25,  # 散点的形状及大小
                   c="red", alpha=0.9,  # 散点的颜色及透明度
                   cmap=None)

    return ax



def get_crop_by_box(x, box_ls, beg_axis=0):
    """
        根据 boxes/box_ls 选定的区域，将 crop_ls 从源张量 x 中截取出来。

        参数：
            x:              <np.array/tensor>
            box_ls:         <list of box>
                                each box is a np.array with shape [batch_size, 2, dimensions]，各个维度的意义为：
                                    2：          box的两个轴对称点
                                    dimensions： 坐标的维度
                                要求：
                                    - 各个 box 应该是已经 sorted 的，亦即小坐标在前大坐标在后。
                                        例如 box=[[1,2],[0,4]] 是错误的。
                                        而 box=[[0,2],[1,4]] 是合法的。
            beg_axis:       <integer> 上面提供的 boxes 中指定的坐标是从 x/crop 的第几个 axis 开始对应的。
                                例如： beg_axis=1 时，box=[[i,j],[m,n]] 表示该 crop 是从原张量的 x[:, i:m, j:n, ...] 部分截取出来的。

        返回：
            crop_ls:        <list of np.array/tensor>
    """
    assert isinstance(box_ls, (list,)) and len(box_ls) > 0
    assert isinstance(beg_axis, (int,)) and 0 <= beg_axis <= x.ndim - box_ls[0].shape[-1], \
        f'0 <= {beg_axis} <= {x.ndim - box_ls[0].shape[-1]} ?'

    # 根据 box 从 x 中截取 crop
    crop_ls = [x[tuple(
        [slice(None, None)] * beg_axis + [slice(beg, end) for beg, end in zip(box[0], box[1])]
    )] for box in box_ls]

    return crop_ls



import pytest
import numpy as np
import math
from kevin_toolbox.math.number_theory import get_greatest_common_divisor
from kevin_toolbox.patches.for_test import check_consistency


def test_get_greatest_common_divisor():
    print("test number_theory.get_greatest_common_divisor()")

    for _ in range(100):
        # 随机构建输入
        n = max(1, int(np.random.rand() * 1000))
        m = max(1, int(np.random.rand() * 1000))
        # 查找
        gcd = get_greatest_common_divisor(n=n, m=m)
        # 标准答案
        gcd_1 = 1
        for i in reversed(range(1, min(n + 1, m + 1))):
            if m % i == 0 and n % i == 0:
                gcd_1 = i
                break
        # 检查
        check_consistency(gcd, gcd_1)



from .cal_iou_between_boxes import cal_iou_between_boxes as cal_iou
from .convert_boxes_from_coord_to_grid_index import \
    convert_boxes_from_coord_to_grid_index as convert_from_coord_to_grid_index
from .get_ticks_of_boxes import get_ticks_of_boxes as get_ticks
from .detect_collision_among_boxes_ls import detect_collision_among_boxes_ls
from .detect_collision_inside_boxes import detect_collision_inside_boxes
from .detect_collision import detect_collision
from .boolean_algebra_for_boxes import boolean_algebra_for_boxes as boolean_algebra
from .cal_area_of_boxes import cal_area_of_boxes as cal_area
from .detect_overlap import detect_overlap



from kevin_toolbox.data_flow.core.reader import Unified_Reader_Base, File_Iterative_Reader


class UReader(Unified_Reader_Base):
    def __index__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_file_iterative_reader(self, file_path, chunk_size, **kwargs):
        return super().get_file_iterative_reader(file_path, chunk_size, **kwargs)

    def get_cache_manager(self, iterator, folder_path):
        return super().get_cache_manager(iterator, folder_path)

    def deal_data(self, data):
        return super().deal_data(data)


class Reader_for_files(UReader):
    """
        从 image_list.txt 中读取 images
            （images 与 features 是一一对应的）

        image_list.txt 的格式为：
            <str of image_path_0>
            ....

        得到的 images 格式为：
            shape = [sample_nums, 1]
            dtype = str
    """

    def get_file_iterative_reader(self, file_path, chunk_size, **kwargs):
        """
            如何迭代读取文件到内存
                要求返回一个按照 chunk_size 来迭代读取文件内容的生成器、迭代器（默认使用 File_Iterative_Reader）
        """
        reader = File_Iterative_Reader(file_path=file_path, chunk_size=chunk_size,
                                       pre_jump_size=kwargs.get("pre_jump_size",None),
                                       jump_size=kwargs.get("jump_size", None),
                                       drop=False, loop_num=1,
                                       convert_func=lambda x: np.array(x, dtype=str))
        return reader


if __name__ == '__main__':
    import numpy as np

    reader = UReader(var=np.ones((10, 1)))

    print(reader.read(5, 10).shape)
    print(reader.read([3, 3]).shape)
    print(reader.find(1))

    reader = UReader(file_path="test_data.txt", chunk_size=2, folder_path="./temp/233")

    print(reader.read(2, 7))
    # del reader

    # reader = UReader(folder_path="./temp/233", chunk_size=2)
    print(len(reader))

    print(reader.read(7, 10))
    print(reader.read([3, 3]))
    print(reader.shape)
    print(len(reader))

    for i in reader:
        print(i)

    print(reader.find('data/6/horse_race_pan/2132020102319002000161_43_4.bmp'))

    reader = Reader_for_files(file_path="test_data.txt", chunk_size=2, pre_jump_size=2, jump_size=2)

    for i in reader:
        print(2333, i)
    print(len(reader))
    # import pdb;
    #
    # pdb.set_trace()
    print(reader.read(100))



from .get_hash import get_hash
from .traverse import traverse
from .get_value_by_name import get_value_by_name
from .set_value_by_name import set_value_by_name
from .count_leaf_node_nums import count_leaf_node_nums
from .get_leaf_nodes import get_leaf_nodes



# from: E:\Temporary_work_area\yolo\code\numerical_characteristics
import numpy as np


def ensure_minmax(bbox):
    """
        输入一个角点形式定义的bbox，转换为 x_1,y_1 是 min，x_2,y_2 是 max 的形式
    """
    x_min, x_max = np.minimum(bbox[..., 0], bbox[..., 2]), np.maximum(bbox[..., 0], bbox[..., 2])
    y_min, y_max = np.minimum(bbox[..., 1], bbox[..., 3]), np.maximum(bbox[..., 1], bbox[..., 3])
    return [x_min, y_min, x_max, y_max]


def trans_minmax_to_xywh(bbox):
    x = (bbox[..., 0] + bbox[..., 2]) / 2
    y = (bbox[..., 1] + bbox[..., 3]) / 2
    w = np.abs(bbox[..., 0] - bbox[..., 2])
    h = np.abs(bbox[..., 1] - bbox[..., 3])
    return [x, y, w, h]


def trans_xywh_to_minmax(bbox):
    x_min = bbox[..., 0] - bbox[..., 2] / 2
    x_max = x_min + bbox[..., 2]
    y_min = bbox[..., 1] - bbox[..., 3] / 2
    y_max = y_min + bbox[..., 3]
    return [x_min, y_min, x_max, y_max]


def cal_IOU(bbox_1, bbox_2, **kwargs):
    """
        选取输入的 bbox_1, bbox_2 中的最后4维来计算 iou
        参数：
            可以通过该mode_1和mode_2参数来指定输入的 bbox_1, bbox_2 是采用哪种形式
            目前支持两种形式：
                角点形式："minmax"
                中心点加长宽的形式："xywh"
    """
    # 参数
    mode_1 = kwargs.get("mode_1", "minmax")
    mode_2 = kwargs.get("mode_2", "minmax")
    assert mode_1 in ("minmax", "xywh")  # 两种模式
    assert mode_2 in ("minmax", "xywh")
    # pre
    # 如果是 xywh 模式，先转换为 minmax 的形式
    x_11, y_11, x_12, y_12 = trans_xywh_to_minmax(bbox_1) if mode_1 == "xywh" else ensure_minmax(bbox_1)
    x_21, y_21, x_22, y_22 = trans_xywh_to_minmax(bbox_2) if mode_2 == "xywh" else ensure_minmax(bbox_2)

    # intersection
    x_1, y_1 = np.maximum(x_11, x_21), np.maximum(y_11, y_21)
    x_2, y_2 = np.minimum(x_12, x_22), np.minimum(y_12, y_22)

    width = x_2 - x_1
    height = y_2 - y_1

    # no intersection when width <= 0 or height <= 0
    width = np.where(width < 0, 0, width)
    height = np.where(height < 0, 0, height)
    intersection = width * height

    # union
    union = (x_12 - x_11) * (y_12 - y_11) + (x_22 - x_21) * (y_22 - y_21)
    union -= intersection

    # iou
    iou = intersection / union

    return iou


if __name__ == '__main__':
    b1, b2 = np.random.rand(4, 5, 6, 10), np.random.rand(4, 5, 6, 10)
    print(cal_IOU(b1, b2).shape)

    b1, b2 = np.array([1, 1, 3, 3]), np.array([2, 4, 4, 2])  # 答案是 1/7
    print(cal_IOU(b1, b2))

    b1, b2 = np.array([1, 1, 3, 3]), np.array([3, 3, 2, 2])  # 答案是 1/7
    mode_2 = "xywh"
    print(cal_IOU(b1, b2, mode_2=mode_2))

    b1, b2 = np.array([1, 1, 3, 3]), np.array([3, 3, 9, 8])  # 答案是 0
    print(cal_IOU(b1, b2))

    b1, b2 = np.array([384, 36, 402, 55]), np.array([385, 49, 402, 65])
    mode_2 = "minmax"
    print(cal_IOU(b1, b2, mode_2=mode_2))
    # 0.19921875



from abc import ABC, abstractmethod
from .sequence_map_base import Sequence_Map_Base
from .iterator_base import Iterator_Base


class My_Iterator_Base(Sequence_Map_Base, Iterator_Base):
    """
        Iterator + Generator 的“pro”版本
            支持 iter(self) 迭代
                self.set_range(beg,end) 设定迭代起点、终点
                self.pass_by(num)    跳过若干个值
            支持 self[index] 取值、len(self) 获取长度属性
    """

    # def __init__(self):
    #     self.beg, self.end, self.index = 0, 0, 0

    @abstractmethod
    def set_range(self, beg, end):
        # self.beg = beg
        # self.end = end
        pass

    @abstractmethod
    def pass_by(self, num):
        # self.index += num
        pass

    @staticmethod
    def round_by_range(beg, end, offset):
        """
            返回 index = beg + offset
        """
        assert isinstance(offset, (int,)), \
            TypeError(f"indices must be integers, not {type(offset)}")
        offset = offset if offset >= 0 else end - beg + offset
        assert 0 <= offset < end - beg, \
            IndexError(f"index {offset} out of range [{0},{end - beg})")
        return beg + offset



import random
import torch
import numpy as np
import copy
from kevin_toolbox.computer_science.algorithm.for_nested_dict_list import set_value_by_name, get_value_by_name


class Strategy_Manager:
    """
        策略管理器

        使用方法：
            # 1.a 初始化
                sm = Strategy_Manager()
            # 2. 添加策略（策略的具体编写方式参见 add() 函数介绍）
                sm.add(
                    strategy={
                        "__dict_form": "para_name:trigger_value",
                        "__trigger_name": "epoch",
                        ":lr": {
                            # 当 key 满足 trigger_value 时，将 ":lr" 指向的部分替换为 value
                            0: 0.1,
                            # 如果是以 <f> 为开头的字符串，则视为函数并将 value 解释为函数执行后的结果
                            #       函数中 t, p 参数将被分别传入 trigger_value 和 para_value
                            "<f>lambda t: t%100==0": "<f>lambda p, t: p*0.1",
                        },
                    },
                    override=False,
                )
            # 3.调用策略计算，得到调整后的结果
                res = sm.cal(
                    trigger_state=dict(epoch=300, step=1, ),
                    ref=dict(lr=5)
                )
                # res: {'lr': 0.5}

            # 1.b 在初始化时添加策略
                sm = Strategy_Manager(
                    strategy=<dict> or <list of dict>,
                    override=False,
                    ...
                )
    """

    def __init__(self, **kwargs):
        """
            参数：
                strategy:           <dict / list of dict> 待添加的多个策略
                                        当其为 dict 时将调用 self.add() 进行添加，
                                        当其为 list 时，且其中元素为 dict 时，将迭代调用 self.add() 添加其中的策略。
                override:           <boolean> 当输入的策略与现有策略发生冲突时，是否强制覆盖旧策略。
                                        将作为 self.add() 中 override 的默认值。
                                        默认为 False
        """
        # 默认参数
        paras = {
            "strategy": tuple(),
            "override": False,
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        if not isinstance(paras["strategy"], (list, tuple,)):
            paras["strategy"] = [paras["strategy"]]
        for v in paras["strategy"]:
            assert isinstance(v, (dict,))
        self.paras = paras

        self.database = dict()
        for strategy in self.paras["strategy"]:
            self.add(strategy=strategy)

    def add(self, strategy, **kwargs):
        """
            将输入的 strategy 策略解释并添加到内部的 database 中

            参数：
                strategy:       <dict of paras> 策略。
                                    支持以下两种组织方式：
                                        （在下面的例子中，我们以 epoch 为触发器，目标变量为 var={"lr":xx, "ratio_ls":[xx,xx]}，
                                        尝试通过策略让变量中的 lr 和 ratio_ls 随着 epoch 而更新。）

                                        1. 触发值作为键。
                                            strategy={
                                                "__dict_form": "trigger_value:para_name",  # 字典的组织方式。必要参数
                                                "__trigger_name": "epoch",  # 触发器的名字。必要参数
                                                0: {
                                                    ":lr": 0.1,
                                                    ":ratio_ls": [1e-3, 1e-2],
                                                },  # 在 epoch=0 时，将 lr 和 ratio_ls 设置为 0.1 和 [1e-3, 1e-2]
                                                300: {
                                                    ":ratio_ls@1": 1e-5,
                                                },  # 在 epoch=300 时，将 ratio_ls[1] 设置为 1e-5
                                                "<f>lambda t: t%100==0": {
                                                    ":lr": "<f>lambda p, t: p*0.1",
                                                },  # 当键为 string 且 开头带有 <f> 标记时候，将以函数的方式读取该字符串，
                                                    # 并在匹配过程中向该函数输入触发值 t 和当前元素的值 p，当函数返回True视为匹配成功。
                                                    # 比如上式表示的是：每经过 100 epoch，也就是当 epoch%100==0 时，lr 在原来的基础上乘上0.1。
                                                    # 函数匹配的的优先级低于直接的值匹配。
                                                ...
                                            }
                                            （不同的触发器可以有不同的触发值类型，不一定是 int 类型，也可以是 float、str 等等）

                                        2. 参数名作为键。
                                            strategy={
                                                "__dict_form": "para_name:trigger_value",
                                                "__trigger_name": "epoch",
                                                ":lr": {
                                                    0: 0.1,
                                                    "<f>lambda t: t%100==0": "<f>lambda p, t: p*0.1",
                                                },
                                                ":ratio_ls": {
                                                    0: [1e-3, 1e-2],
                                                },
                                                ":ratio_ls@1": {
                                                    300: 1e-5,
                                                },
                                                ...
                                            }
                                        以上两种形式等效。

                                        其中诸如 ":ratio_ls@1" 的参数名的含义参见 computer_science.algorithm.utils 下的
                                        set_value_by_name() 函数介绍

                override:       <boolean> 当输入的策略与现有策略发生冲突时，是否强制覆盖旧策略。
                                    默认为 False，在发生冲突时候将直接抛出错误。
        """
        # 校验参数
        assert strategy.get("__dict_form", None) in ["trigger_value:para_name", "para_name:trigger_value"]
        assert isinstance(strategy.get("__trigger_name", None), (str,))
        strategy = copy.deepcopy(strategy)
        override = kwargs.get("override", self.paras["override"])

        _dict_form, _trigger_name = strategy.pop("__dict_form"), strategy.pop("__trigger_name")

        # 将 2. 参数名作为键 的形式转换为 1. 触发值作为键 的形式
        if _dict_form == "para_name:trigger_value":
            temp = dict()
            for p_key in strategy.keys():
                assert isinstance(p_key, (str,))
                for t_key in strategy[p_key].keys():
                    if t_key not in temp:
                        temp[t_key] = dict()
                    temp[t_key][p_key] = strategy[p_key][t_key]
            strategy = temp

        # 将策略添加到 database
        old_strategy = self.database.get(_trigger_name, dict())
        for t_key, item in strategy.items():
            if t_key not in old_strategy:
                old_strategy[t_key] = dict()
            #
            if not override:
                # 检查冲突
                temp = set(old_strategy[t_key].keys()).intersection(set(item.keys()))
                assert len(temp) == 0, \
                    f'conflicting strategies for trigger {_trigger_name} in value {t_key}: \n' \
                    f'\tpara_name: {temp}'
            #
            old_strategy[t_key].update(item)

        self.database[_trigger_name] = old_strategy

    def cal(self, var, trigger_state):
        """
            从 database 中读取 trigger_state 指定的触发器状态下的策略，并根据该策略修改目标变量 var 中的对应部分

            参数：
                var:                    <list/dict> 目标变量
                trigger_state:          <dict> 触发器状态

            返回:
                var, action_s_all
                    其中 action_s_all 是一个以 trigger_name 为键的 dict，下面各值为在该触发器状态下匹配上的策略
        """
        assert isinstance(trigger_state, (dict,)) and isinstance(var, (dict, list,))

        # 查找策略
        action_s_all = dict()
        for t_name, t_value in trigger_state.items():
            if t_name not in self.database:
                continue
            strategy = self.database[t_name]
            #
            action_s = dict()  # {p_name: p_value, ...}
            # 使用匹配函数
            for key, p_s in strategy.items():
                if isinstance(key, (str,)) and key.startswith("<f>") and eval(key[3:])(t_value):
                    action_s.update({i: j for i, j in p_s.items() if i not in action_s})
            # 直接匹配
            if t_value in strategy:
                action_s.update(strategy[t_value])
            # !! 这一步排序很关键，它保证了将先处理整体，然后再处理局部
            #   比如对于 ':ratio_ls' 和 ':ratio_ls@0'，将先处理前者
            action_s_all[t_name] = [(k, action_s[k]) for k in sorted(action_s.keys())]

        # 执行策略
        for t_name, action_s in sorted(action_s_all.items(), key=lambda x: x[0]):
            t_value = trigger_state[t_name]
            for name, p_value in copy.deepcopy(action_s):
                if isinstance(p_value, (str,)) and p_value.startswith("<f>"):
                    raw_value = get_value_by_name(var=var, name=name)
                    p_value = eval(p_value[3:])(raw_value, t_value)
                set_value_by_name(var=var, name=name, value=p_value)

        return var, action_s_all

    def __str__(self):
        return str(self.database)


if __name__ == '__main__':
    sm = Strategy_Manager(strategy={
        "__dict_form": "para_name:trigger_value",
        "__trigger_name": "epoch",
        ":lr": {
            # 0: 0.1,
            "<f>lambda t: t%100==0": "<f>lambda p, t: p*0.1",
        },
    })
    sm.add(strategy={
        "__dict_form": "trigger_value:para_name",
        "__trigger_name": "epoch",
        0: {
            ":lr": 0.1,
            ":ratio_ls": [1e-3, 1e-2],
        },
        300: {
            ":ratio_ls@1": 1e-5,
        },
        # "<f>lambda t: t%100==0": {
        #     ":lr": "<f>lambda p, t: p*0.1",
        # },
    }, override=False)

    print(sm.database)

    var, action_s_all = sm.cal(trigger_state=dict(epoch=0, ), var=dict())
    print(var, action_s_all)

    var, action_s_all = sm.cal(trigger_state=dict(epoch=300, ), var=var)
    print(var, action_s_all)



def traverse(var, match_cond, action_mode="remove", converter=None, b_use_name_as_idx=False):
    """
        遍历 var 找到符合 match_cond 的元素，将其按照 action_mode 指定的操作进行处理

        参数：
            var:                待处理数据
                                    当 var 不是 dict 或者 list 时，将直接返回 var 而不做处理
            match_cond:         <func> 元素的匹配条件
                                    函数类型为 def(parent_type, idx, value): xxxx
                                    其中：
                                        parent_type     该元素源自哪种结构体，有两个可能传入的值： list，dict
                                        idx             该元素在结构体中的位置
                                                            当 b_use_name_as_idx=False 时，
                                                                对于列表是 index，对于字典是 key
                                                            当为 True 时，传入的是元素在整体结构中的 name 位置，name的格式和含义参考
                                                                get_value_by_name() 中的介绍
                                        value           元素的值
            action_mode:        <str> 如何对匹配上的元素进行处理
                                    目前支持：
                                        "remove"        将该元素移除
                                        "replace"       将该元素替换为 converter() 处理后的结果
                                        "skip":          不进行任何操作
            converter:          <func> 参见 action_mode 中的 "replace" 模式
                                    函数类型为 def(idx, value): xxxx
                                    其中 idx 和 value 的含义参见参数 match_cond 介绍
            b_use_name_as_idx     <boolean> 对于 match_cond/converter 中的 idx 参数，是传入整体的 name 还是父节点的 index 或 key。
                                    默认为 False
    """
    assert callable(match_cond)
    assert action_mode in {"replace", "remove", "skip"}
    if action_mode == "replace":
        assert callable(converter)

    return recursive_(var, match_cond, action_mode, converter, b_use_name_as_idx, "")


def recursive_(var, match_cond, action_mode, converter, b_use_name_as_idx, pre_name):
    if isinstance(var, (list, dict)):
        items = reversed(list(enumerate(var))) if isinstance(var, list) else list(var.items())
        for k, v in items:
            if b_use_name_as_idx:
                idx = f'{pre_name}@{k}' if isinstance(var, list) else f'{pre_name}:{k}'
            else:
                idx = k
            if match_cond(type(var), idx, v):
                if action_mode == "remove":
                    var.pop(k)
                elif action_mode == "replace":
                    var[k] = converter(idx, v)
                else:
                    pass
            else:
                var[k] = recursive_(v, match_cond, action_mode, converter, b_use_name_as_idx, idx)
    else:
        pass
    return var



import torch
from kevin_toolbox.env_info import version

"""
    在 1.8 及其以下版本的 pytorch 中，没有 torch.linalg.svd()
        改用 torch.svd() 来实现，
        实现过程中考虑到两者的输出的 V 矩阵发生了转置，同时输入的 full_matrices 与 some 互为反义
"""

if version.compare(torch.__version__, "<=", "1.8", mode="short"):
    @torch.no_grad()
    def svd(X, full_matrices=True):
        U, S, V = torch.svd(X, some=not full_matrices)
        V = V.transpose(-2, -1)
        return U, S, V
else:
    svd = torch.linalg.svd



from kevin_toolbox.computer_science.algorithm.for_nested_dict_list import traverse


def count_leaf_node_nums(var):
    """
        获取嵌套字典列表 var 中所有叶节点的数量

        参数：
            var:                待处理数据
                                    当 var 不是 dict 或者 list 时，直接将 var 视为根节点，叶节点数量视为 0
    """
    nums = 0

    def count(_, __, v):
        nonlocal nums
        if not isinstance(v, (list, dict,)):
            nums += 1
        return False

    traverse(var=var, match_cond=count, action_mode="skip", b_use_name_as_idx=False)
    return nums



import numpy as np
from kevin_toolbox.computer_science.algorithm.combinatorial_optimization import zero_one_knapsack_problem


def get_subset_with_largest_product(ls, upper_bound):
    """
        找出乘积不大于 upper_bound 的 ls 的最大乘积子集
            要求 ls 中的元素，以及 upper_bound 都为正数

        参数：
            ls:             <list>
            upper_bound:    <int/float>

        返回：
            product, subset
            最大子集的乘积 ， 最大子集
                当解不存在时候，返回 None, None
    """
    assert isinstance(ls, (list, tuple,)) and len(ls) > 0
    for i in ls + [upper_bound]:
        assert isinstance(i, (int, float, np.number,)) and i > 0

    _, idx_ls = zero_one_knapsack_problem(weights=[np.log(i) for i in ls],
                                          upper_bound=np.log(upper_bound + 1e-5))  # 为避免浮点数精度问题，需要为 upper_bound 添加冗余
    if idx_ls is not None and np.prod([ls[i] for i in idx_ls]) > upper_bound:
        # 如果前面添加的冗余影响了结果，导致乘积实际上已经超过了 upper_bound
        #   那么就尝试不添加冗余重新计算
        _, idx_ls = zero_one_knapsack_problem(weights=[np.log(i) for i in ls], upper_bound=np.log(upper_bound))

    if idx_ls is not None:
        subset = [ls[i] for i in idx_ls]
        product = np.prod(subset)
    else:
        product, subset = None, None
    return product, subset


if __name__ == '__main__':
    print(get_subset_with_largest_product(ls=[1.2, 3.4], upper_bound=12))
    print(get_subset_with_largest_product(ls=[1, 2, 2, 2, 2, 2, 2, 2, 2, 2], upper_bound=100))
    print(get_subset_with_largest_product(ls=[6, 15, 2, 2, 9], upper_bound=18))
    # (4.08, [1.2, 3.4])
    # (64, [2, 2, 2, 2, 2, 2])



import numpy as np


def generate_dct_trans_matrix(**kwargs):
    """
        生成用于进行1维离散余弦变换（DCT）的变换基

        使用方法：
            假设要变换的1维信号队列为 X [k, n]
                其中：
                - n 为信号序列的长度（在DCT中一般将输入的信号序列视为经过时轴对称延拓后得到的周期为2n的序列）
                - k 为信号的通道数。
                你可以将 X 视为 k 个长度为 n 的1维信号的组合。
            使用该函数生成一个转换基 B [n, m]
                其中：
                - m 表示基向量/基函数的数量（数量越大越能整合高频信号）
                - n 为基向量的长度/基函数的离散采样点数量，与输入周期信号的周期的一半相等
            则变换过程为 Y = X @ B
                得到的 Y [k, m]

        如何推广到多维？
            原理：
                由于频域变换的维度可分离性，因此可以将多维 DCT 变换分解为对信号的每个维度单独做1维 DCT 变换。
            具体方法：
                以 2d DCT 变换为例，假设输入信号为 X [k, n0, n1]
                    1.0 首先使用该函数生成针对于维度 n1 的变换基 B1 [n1, m1]
                    1.1 对维度 n1 进行变换：Z = X @ B1，得到 Z [k, n0, m1]
                    1.2 对 Z 进行转置 Z = Z.permute(0, 2, 1) 得到 Z [k, m1, n0]
                    2.0 类似地生成变换基  B0 [n0, m0]
                    2.1 对维度 n0 进行变换：Y = Z @ B0，得到 Y [k, m1, m0]
                    2.2 对 Y 进行转置恢复维度顺序 Y = Y.permute(0, 2, 1) 得到 Z [k, m0, m1]
        参数：
            sampling_points_num:    <integer> 转换矩阵的行数，对应 基函数的离散采样点数量
                                                与输入周期信号的周期的一半相等
            basis_series_num:       <integer> 转换矩阵的列数，对应 基向量/基函数的数量
                                                数量越大越能整合高频信号，但不应超过采样点的数量 sampling_points_num
                                                如果超过则会导致列向量不再两两正交，也不一定保证单位化
            shape:                  <list of integers> 长度为 2 的列表，记录了 [sampling_points_num, basis_series_num]
                当 sampling_points_num ... 和 shape 被同时设定时，以前者为准。

        返回：
            B       <np.array> shape [r_num, c_num]
                矩阵中各元素为
                    B[r,c] := g(c) * sqrt(2/r_num) * cos( (2*r + 1) * c * pi / (2*r_num) )
                        其中 g(c) := sqrt(1/2) if c==0 else 1

        技巧：
            当两个转换矩阵的 r_num 相同时，小矩阵可以直接从大矩阵中截取，而不需要重新计算。
    """
    # 默认参数
    paras = {
        # 必要参数
        "sampling_points_num": None,
        "basis_series_num": None,
        # 别名
        "shape": None,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    for i, k in enumerate(["sampling_points_num", "basis_series_num"]):
        if paras[k] is None:
            assert isinstance(paras["shape"], (list, tuple,)) and len(paras["shape"]) == 2
            paras[k] = paras["shape"][i]
        assert isinstance(paras[k], (int,)) and paras[k] > 0
    r_num, c_num = paras["sampling_points_num"], paras["basis_series_num"]
    if c_num > r_num:
        print(f"Warning: basis_series_num {c_num} should not be larger than sampling_points_num {r_num}!")

    B = np.zeros(shape=(r_num, c_num))
    # cos
    for r in range(r_num):
        for c in range(c_num):
            B[r, c] = np.cos((2 * r + 1) * c * np.pi / (2 * r_num))

    # g(c) * sqrt(2/r_num)
    g = np.ones(shape=(1, c_num)) * np.sqrt(2 / r_num)
    g[..., 0] = g[..., 0] * np.sqrt(1 / 2)
    B = B * g

    return B


if __name__ == '__main__':
    print(generate_dct_trans_matrix(shape=[4, 4]))
    print(generate_dct_trans_matrix(shape=[6, 4]))
    print(generate_dct_trans_matrix(shape=[4, 6]))
    print(np.sum(generate_dct_trans_matrix(shape=[1, 2]) ** 2, axis=0))



import numpy as np
from kevin_toolbox.math.number_theory import get_primes


def prime_factorization(n):
    """
        对正整数n进行质因数分解
            返回它的所有素数因子，包括1
    """
    assert isinstance(n, (int, np.integer,)) and n >= 1

    primes = get_primes(n=n + 1)
    factors = [1]
    for p in primes[1:]:
        while n % p == 0:
            factors.append(p)
            n /= p
        if p > n:
            break
    assert n == 1

    return factors



import os
import numpy as np
from kevin_toolbox.geometry import for_boxes
from kevin_toolbox.math.dimension import coordinates

U_OPERATION_CHOICES = {None, "not"}
BI_OPERATION_CHOICES = {'or', 'and', 'diff'}


def boolean_algebra_for_boxes(**kwargs):
    """
        布尔运算
            对 boxes_ls 中的各个 boxes，按照 binary_operation_ls、unary_operation_ls 中指定的操作进行布尔运算

        参数：
            boxes_ls:          <list of boxes/None>
                                where boxes is <np.array> with shape [batch, 2, dimensions]
                                各个维度的意义为：
                                    batch：      box的数量
                                    2：          box的两个轴对称点
                                    dimensions： 坐标的维度
                                where None represents the empty set
            binary_operation_ls:    <list of string> 二元运算操作（对两个相邻的 box进行操作）
                                支持以下运算符：
                                    "and":      与
                                    "or":       或
                                    "diff":     减去， a diff b 等效于 a and (not b)
                                注意：
                                    - 因为二元运算符是对两个 box 进行操作的，因此 binary_operation_ls 的长度需要比 boxes_ls 小 1
            unary_operation_ls:     <list> 一元运算符
                                支持以下运算符：
                                    "not":      取反
                                    None:       不进行运算
                                默认为 None，表示不进行任何一元运算
                                注意：
                                    - 当 unary_operation_ls 设定有具体值时，要求其长度与 boxes_ls 相等
                                    - ！！当使用 "not" 运算时，默认使用 boxes_ls 的最小外切长方体作为全集。
                                        如果要指定全集 U 的范围，建议在第一个元素 a 前添加操作 U and a，
                                        该操作将显式地声明全集范围。
        返回：
            boxes 当结果为空集时，返回 None
    """
    # 默认参数
    paras = {
        # 必要参数
        "boxes_ls": None,
        "binary_operation_ls": None,
        "unary_operation_ls": None,
    }

    # 获取参数
    paras.update(kwargs)
    global U_OPERATION_CHOICES, BI_OPERATION_CHOICES

    # 校验参数
    assert isinstance(paras["boxes_ls"], (list,))
    boxes_ls = paras["boxes_ls"]
    if len(boxes_ls) == 0:
        return None
    # ticks
    temp = [boxes for boxes in boxes_ls if boxes is not None]
    if len(temp) == 0:
        return None
    elif len(temp) > 1:
        temp = np.concatenate(temp, axis=0).astype(dtype=np.float32)
    else:
        temp = temp[0]
    ticks = for_boxes.get_ticks(boxes=temp)
    #
    if paras["unary_operation_ls"] is None:
        paras["unary_operation_ls"] = [None] * len(boxes_ls)
    paras["binary_operation_ls"].insert(0, "or")  # 添加哨兵
    #
    for key, choices in zip(["unary_operation_ls", "binary_operation_ls"], [U_OPERATION_CHOICES, BI_OPERATION_CHOICES]):
        assert len(paras[key]) == len(boxes_ls)
        assert set(paras[key]).issubset(choices), \
            f'ele in {key} should be string in {choices}, but get {set(paras[key])}'
    unary_operation_ls = paras["unary_operation_ls"]
    binary_operation_ls = paras["binary_operation_ls"]

    # convert to grid
    settings_for_grid = dict(mode="closed", ticks=ticks)
    grids_ls = []
    for boxes in boxes_ls:
        if boxes is None:
            grids_ls.append(None)
        else:
            grids_ls.append(
                for_boxes.convert_from_coord_to_grid_index(boxes=boxes, settings_for_grid=settings_for_grid))

    # convert to matrix
    grid_shape = [len(tk_ls) - 1 for tk_ls in ticks]
    res_matrix = np.zeros(shape=grid_shape, dtype=bool)
    for i, grids in enumerate(grids_ls):
        bi_operation = binary_operation_ls[i]
        u_operation = unary_operation_ls[i]
        #
        matrix = np.zeros(shape=grid_shape, dtype=bool)
        if grids is not None:
            for grid in grids:
                scope = ','.join([f"{int(beg)}:{int(end)}" for beg, end in zip(grid[0], grid[1])])
                exec(f"matrix[{scope}]=1")
        #
        if u_operation is None:
            pass
        elif u_operation == "not":
            matrix = ~matrix
        else:
            raise ValueError
        #
        if bi_operation == "or":
            res_matrix = res_matrix | matrix
        elif bi_operation == "and":
            res_matrix = res_matrix & matrix
        elif bi_operation == "diff":  # a diff b == a and (not b)
            res_matrix = res_matrix & ~matrix
        else:
            raise ValueError

    # pick result from matrix
    zip_indices = np.where(res_matrix)
    indices_ls = coordinates.convert(var=zip_indices, shape=res_matrix.shape,
                                     input_format="zip_indices", output_format="indices_ls")
    # convert back to box_ls
    res_boxes = None
    if len(indices_ls) > 0:
        beg = indices_ls.reshape(indices_ls.shape[0], 1, -1)
        end = beg + 1
        res_grids = np.concatenate([beg, end], axis=1)
        res_boxes = for_boxes.convert_from_coord_to_grid_index(boxes=res_grids, settings_for_grid=settings_for_grid,
                                                               reverse=True)
    return res_boxes



from kevin_toolbox.computer_science.data_structure import Executor
from kevin_toolbox.machine_learning.dataset.face.verification.merge import merge


def get_executor_ls_by_block_along_axis(factory, i_0, i_1, j_0, j_1, axis_to_split, size_upper, need_to_generate,
                                        pre_executor=None, pre_size=None):
    """
        通过调用 verification.Factory 中的 generate_by_block() 函数，
            来在（i_0, i_1, j_0, j_1）指定的子矩阵范围内，
            沿着 axis 指定的轴线方向，以 size_upper 为目标大小进行分割，
            生成一系列的执行器 executor_ls，
            每个执行器在被 executor() 调用后都将返回一个数据集
        参数：
            factory:                verification.Factory 实例
            i_0, i_1, j_0, j_1:     子矩阵的范围
                                        （支持多种方式输入，具体参见 Factory.generate_by_block() 中的介绍）
            axis_to_split:          分割子矩阵的方向
                                        可选值： ["i", "j"]
                                        以 "i" 为例，此时子矩阵沿行分割为 [i_0, i_0+step, ... ,i_1]
            size_upper:             每个分块的目标大小（大小的上界）
            pre_executor:           前置的数据集生成器
                                        第一个chunk产生的数据集将与pre_executor整合为一个dataset并返回
            pre_size:               前置数据集的大小
            need_to_generate:       需要生成的字段
                                        （参见 Factory.generate_by_block() 中的介绍）

        关于返回的数据集大小（以axis="i"为例，令j_len:=j_1-j_0）：
            当 pre_executor=None 时，
                对于前面完整的矩形，该矩形的行数将根据与size_upper相差的部分的数量计算得到，
                从而保证合并后返回的数据集尽量贴近上界，有
                dataset_size = size_upper//j_len * j_len
                对于最后一个残缺的矩形，有
                dataset_size <= size_upper
            当 pre_executor 有值时，
                第一个矩形将与 pre_executor 整合为一个dataset，有
                dataset_size <= min( size_upper//j_len * j_len, len(pre_executor) )
                其他同上。
    """
    assert axis_to_split in ["i", "j"]
    i_len = i_1 - i_0 if isinstance(i_0, (int,)) else len(i_0)
    j_len = j_1 - j_0 if isinstance(j_0, (int,)) else len(j_0)
    assert i_len > 0 and j_len > 0
    len_fix, len_to_split = (j_len, i_len) if axis_to_split == "i" else (i_len, j_len)

    executor_ls, size_ls = [], []
    #
    count = 0
    while count < len_to_split:
        # step
        if count == 0 and pre_executor is not None:
            # 对于第一个矩形
            assert isinstance(pre_size, (int,))
            step = (size_upper - pre_size) // len_fix
            if step <= 0:  # pre_executor 已经够大了，不需要再补充
                executor_ls.append(pre_executor)
                pre_executor = None
                continue
        else:
            step = size_upper // len_fix
        step = min(step, len_to_split - count)

        # 范围
        if axis_to_split == "i":
            i_0_, i_1_ = count, count + step
            if not isinstance(i_0, (int,)):
                i_0_, i_1_ = i_0[i_0_:i_1_], None
            j_0_, j_1_ = j_0, j_1
        else:
            j_0_, j_1_ = count, count + step
            if not isinstance(j_0, (int,)):
                j_0_, j_1_ = j_0[j_0_:j_1_], None
            i_0_, i_1_ = i_0, i_1

        # 计算
        paras = dict(i_0=i_0_, i_1=i_1_, j_0=j_0_, j_1=j_1_,
                     pick_triangle=False, need_to_generate=need_to_generate)
        executor_ls.append(Executor(func=factory.generate_by_block,
                                    kwargs=paras))
        size_ls.append(factory.cal_size_of_block(**paras))

        # 对于第一个矩形，进行合并
        if count == 0 and pre_executor is not None:
            # 数据集
            u, v = pre_executor, executor_ls.pop(0)
            executor_ls.insert(0, Executor(func=merge, f_args=[u, v]))
            # size
            u, v = pre_size, size_ls.pop(0)
            size_ls.insert(0, u + v)

        #
        count += step

    return executor_ls, size_ls



from kevin_toolbox.computer_science.data_structure import Executor
from .along_axis import get_executor_ls_by_block_along_axis


def get_executor_ls_by_block_of_all(factory, chunk_step, need_to_generate, **kwargs):
    """
        通过调用 verification.Factory 中的 generate_by_block() 函数，
            来对整个矩阵，
            生成一系列的执行器 executor_ls，
            每个执行器在被 executor() 调用后都将返回一个数据集
        参数：
            factory:                verification.Factory 实例
            chunk_step:             每个分块的大小
            need_to_generate:       需要生成的字段
                                        （参见 Factory.generate_by_block() 中的介绍）
    """

    width = len(factory.paras["features"])  # 矩阵的宽度
    assert width > 0, \
        Exception("Error: the length of features in the input factory should be larger than 0!")

    # block的数量
    chunk_nums = (width - 1) // chunk_step + 1

    """
    执行器
    """
    executor_ls, size_ls = [], []

    # 对于完整的 block（去除最后一行、列上的block）
    #     将整个block作为一个dataset
    #     dataset_size = chunk_step * chunk_step
    for i in range(chunk_nums - 1):
        i_0 = i * chunk_step
        i_1 = i_0 + chunk_step
        for j in range(chunk_nums - 1):
            j_0 = j * chunk_step
            j_1 = j_0 + chunk_step
            # 计算
            paras = dict(i_0=i_0, i_1=i_1, j_0=j_0, j_1=j_1,
                         pick_triangle=False, need_to_generate=need_to_generate)
            executor_ls.append(Executor(func=factory.generate_by_block,
                                        kwargs=paras))
            size_ls.append(factory.cal_size_of_block(**paras))

    # 对于最后一行
    #     对于完整的矩形有
    #     chunk_step * chunk_step < dataset_size <= chunk_step * (chunk_step + 1)
    # 注意：
    #     对于最后一个可能残缺的矩阵（亦即大小不满足下界），
    #     该部分将与接下来的最后一列的部分头部组成一个dataset
    executor_ls_temp, size_ls_temp = get_executor_ls_by_block_along_axis(factory,
                                                                         i_0=(chunk_nums - 1) * chunk_step,
                                                                         i_1=width,
                                                                         j_0=0,
                                                                         j_1=width,
                                                                         axis_to_split="j", size_upper=size_upper,
                                                                         need_to_generate=need_to_generate)
    executor_ls.extend(executor_ls_temp)
    size_ls.extend(size_ls_temp)

    # 对于最后一列
    #     第一个矩形将与前面残缺的部分整合为一个dataset，该矩形的行数将根据缺少部分的数量计算得到，从而保证
    #     chunk_step * chunk_step <= dataset_size <= chunk_step * (chunk_step + 1)
    #     接下来的矩形将调整行数以尽量贴近上界，
    #     对于完整的矩形有
    #     chunk_step * chunk_step < dataset_size <= chunk_step * (chunk_step + 1)
    #     对于最后一个可能残缺的矩阵，有
    #     dataset_size <= chunk_step * (chunk_step + 1)
    # 这里承接上面的  res
    if chunk_nums > 1:
        executor_ls_temp, size_ls_temp = get_executor_ls_by_block_along_axis(factory,
                                                                             i_0=0,
                                                                             i_1=(chunk_nums - 1) * chunk_step,
                                                                             j_0=(chunk_nums - 1) * chunk_step,
                                                                             j_1=width,
                                                                             axis_to_split="i", size_upper=size_upper,
                                                                             need_to_generate=need_to_generate,
                                                                             pre_executor=executor_ls.pop(-1),
                                                                             pre_size=size_ls.pop(-1))
        executor_ls.extend(executor_ls_temp)
        size_ls.extend(size_ls_temp)

    # 综合而言，
    #     除最后一个dataset以外，都有
    #     chunk_step * chunk_step <= dataset_size <= chunk_step * (chunk_step + 1)
    #     最后一个dataset可能是残缺的，有
    #     0 < dataset_size <= chunk_step * (chunk_step + 1)
    return executor_ls, size_ls



from .along_axis import get_executor_ls_by_block_along_axis as along_axis
from .along_diagonal import get_executor_ls_by_block_along_diagonal as along_diagonal
from .of_all import get_executor_ls_by_block_of_all as of_all
from .of_triangle import get_executor_ls_by_block_of_triangle as of_triangle



import numpy as np


def cal_area_of_boxes(boxes, is_sorted=True):
    """
        计算体积
    """
    if boxes is not None:
        if not is_sorted:
            boxes = np.sort(boxes, axis=1)
        area = (boxes[:, 1] - boxes[:, 0]).prod(axis=-1).sum()
    else:
        area = 0
    return area



def set_crop_by_box(x, box_ls, crop_ls, beg_axis=0):
    """
        将 crop_ls 填充到 x 中 boxes 指定的区域
            将直接在输入的 x 上进行 inplace 赋值操作

        参数：
            x:              <np.array/tensor>
            box_ls:         <list of box>
                                each box is a np.array with shape [batch_size, 2, dimensions]，各个维度的意义为：
                                    2：          box的两个轴对称点
                                    dimensions： 坐标的维度
                                要求：
                                    - 各个 box 应该是已经 sorted 的，亦即小坐标在前大坐标在后。
                                        例如 box=[[1,2],[0,4]] 是错误的。
                                        而 box=[[0,2],[1,4]] 是合法的。
            crop_ls:        <list of np.array/tensor> 需要与 boxes 一一对应
            beg_axis:       <integer> 上面提供的 boxes 中指定的坐标是从 x/crop 的第几个 axis 开始对应的。
                                例如： beg_axis=1 时，box=[[i,j],[m,n]] 表示该 crop 是从原张量的 x[:, i:m, j:n, ...] 部分截取出来的。

        返回：
            x:              <np.array/tensor>
    """
    assert isinstance(box_ls, (list,)) and len(box_ls) > 0
    assert isinstance(beg_axis, (int,)) and 0 <= beg_axis <= x.ndim - box_ls[0].shape[-1], \
        f'0 <= {beg_axis} <= {x.ndim - box_ls[0].shape[-1]} ?'
    assert isinstance(crop_ls, (list,)) and len(crop_ls) == len(box_ls)

    for crop, box in zip(crop_ls, box_ls):
        x[tuple([slice(None, None)] * beg_axis + [slice(beg, end) for beg, end in zip(box[0], box[1])])] = crop

    return x



def convert_dict_key_to_number(x):
    """
        尝试将字典中的所有 key 转换为数字
    """
    if isinstance(x, dict):
        res = dict()
        for k, v in x.items():
            try:
                k = int(k)
            except:
                try:
                    k = float(k)
                except:
                    pass
            res[k] = v
        return res
    else:
        return x


if __name__ == '__main__':
    print(convert_dict_key_to_number({"123": 123, "-1": -1, "1.23": 1.23, "-1.000": -1.0, "1.2.3": None}))



import torch
import numpy as np

FUNCTION_TABLE_GALLERY = dict(
    np_array=dict(
        swapaxes=lambda x, axis_0, axis_1: np.swapaxes(x, axis_0, axis_1),
        permute=lambda x, axis_ls: np.transpose(x, axis_ls),
        concat=lambda x_ls, axis: np.concatenate(x_ls, axis=axis),
        flatten=lambda x, axis_0, axis_1: x.reshape(list(x.shape[:axis_0]) + [-1] + list(x.shape[axis_1 + 1:]))
    ),
    torch_tensor=dict(
        swapaxes=lambda x, axis_0, axis_1: torch.transpose(x, axis_0, axis_1),
        permute=lambda x, axis_ls: x.permute(*axis_ls),
        concat=lambda x_ls, axis: torch.cat(x_ls, dim=axis),
        flatten=lambda x, axis_0, axis_1: torch.flatten(x, start_dim=axis_0, end_dim=axis_1),
    )
)
"""
已注册函数及其使用方法：
    - swapaxes(x, axis_0, axis_1)
    - permute(x, axis_ls)
    - concat(x_ls, axis)
    - flatten(x, axis_0, axis_1)  将 axis_0 到 axis_1 之间的轴进行展平
"""


def get_function_table_for_array_and_tensor(x):
    """
        根据输入 x 的类型获取对应的 function_table
            目前 function_table 已覆盖的函数有：
                swapaxes(x, dim0, dim1)  交换两个维度
                permute(x, dim_ls)  对维度进行重排

        返回：
            [type], [function_table]
    """
    if type(x) is np.ndarray:
        key = "np_array"
    elif torch.is_tensor(x):
        key = "torch_tensor"
    else:
        raise ValueError(f'x should be np.array or torch.tensor, but get a {type(x)}')
    return key, FUNCTION_TABLE_GALLERY[key]



import torch
import torch.nn as nn


class Lambda_Layer(nn.Module):
    def __init__(self, func):
        super(Lambda_Layer, self).__init__()
        self.func = func

    def forward(self, x):
        return self.func(x)



class Trigger:
    def __init__(self):
        """
            触发器
                通过 self.update() 来接收并监视 state_dict，
                当 state_dict 全部或部分发生变化时，将会把【发生变化的部分】
                传入绑定的函数中，并执行一次该函数。
        """

        self.last_state = dict()
        self.func_ls = []

    def bind(self, target):
        """
            绑定函数

            参数：
                target:     <func/list of func>
        """
        if not isinstance(target, (list, tuple,)):
            target = [target]
        for func in target:
            assert callable(func)
        self.func_ls.extend(target)

    def update(self, cur_state):
        """
            更新状态，决定是否触发
        """
        assert isinstance(cur_state, (dict,))

        new_state = dict()
        for key, value in cur_state.items():
            if key not in self.last_state or self.last_state[key] != value:
                new_state[key] = value

        if len(new_state) > 0:
            for func in self.func_ls:
                func(new_state)
            self.last_state.update(new_state)

        return new_state



import numpy as np
from .computational_tree import Node


def concat_crops_into_whole(**kwargs):
    """
        将 crop 按照对应的 box 指示的位置，以行优先的内存顺序，进行拼接、展平

        工作流程：
            首先根据 box_ls 构建 computational_tree，然后将 crop 按照对应的 box 分配到计算图中对应的叶节点，调用计算图中的 concat()
            进行合并，最后从计算图的根节点中取出最后合并得到的结果。

        参数：
            crop_ls:        <list of np.array/tensor>
            box_ls:         <list of np.arrays>
                                each element is an array with shape [2, dimensions]
                                各个维度的意义为：
                                    2：          box 的两个轴对称点
                                    dimensions： 坐标的维度
                                要求：
                                    - 各个 box 应该是已经 sorted 的，亦即小坐标在前大坐标在后。
                                        例如 box=[[1,2],[0,4]] 是错误的。
                                        而 box=[[0,2],[1,4]] 是合法的。
                                    - 各个 box 在坐标轴上的投影之间没有重叠部分
                                    函数 geometry.for_boxes.boolean_algebra() 返回的 boxes 结果，以及
                                    函数 geometry.for_boxes.detect_collision() 返回的 node_ls 中每个 node 下面
                                    的 node.description["by_boxes"] 都符合该要求。
            beg_axis:       <integer> 上面提供的 box 中指定的坐标是从 crop 的第几个 axis 开始对应的。
                                例如： beg_axis=1 时，box=[[i,j],[m,n]] 表示该 crop 是从原张量的 [:, i:m, j:n, ...] 部分截取出来的。
            computational_tree: <Node> 计算图
                                默认为 None，函数将根据输入的 box_ls 自动构建计算图。
                                你也可以将已有的计算图代入该参数中，以跳过构建计算图的步骤，节省计算量。
            return_details: <boolean> 是否以详细信息的形式返回结果
                                默认为 False，此时返回：
                                    whole:  <np.array/tensor> 对 crop_ls 进行合并后的结果
                                当设置为 True，将返回一个 dict：
                                    details = dict(
                                        whole = <np.array/tensor>,  # 对 crop_ls 进行合并后的结果
                                        box_ls = <list of np.arrays>,  # 按照 内存顺序 对 box_ls 进行排序后的结果
                                        crop_ls = <list of np.array/tensor>,  # 按照 内存顺序 对 crop_ls 进行排序后的结果
                                        beg_axis = beg_axis,  # 对应与输入的 beg_axis
                                        computational_tree = <Node>,  # 计算图
                                    )
        返回：
            whole 或者 details
    """
    # 默认参数
    paras = {
        # 必要参数
        "crop_ls": None,
        "box_ls": None,
        #
        "beg_axis": 0,
        "computational_tree": None,
        "return_details": False,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    assert isinstance(paras["crop_ls"], (list,)) and len(paras["crop_ls"]) > 0
    crop_ls = paras["crop_ls"]
    #
    assert isinstance(paras["box_ls"], (list,)) and len(paras["box_ls"]) == len(crop_ls)
    assert paras["box_ls"][0].ndim == 2 and paras["box_ls"][0].shape[1] == 2
    box_ls = paras["box_ls"]
    #
    assert isinstance(paras["beg_axis"], (int,))
    beg_axis = paras["beg_axis"]
    end_axis = beg_axis + box_ls[0].shape[-1] - 1
    assert 0 <= beg_axis <= end_axis < crop_ls[0].ndim
    # 构建计算图
    if paras["computational_tree"] is None:
        tree = Node(box_ls=box_ls)
        tree.build_tree()
        tree.init_tree()
    else:
        tree = paras["computational_tree"]
    assert isinstance(tree, (Node,))

    # 按行优先进行多级排序
    sorted_node_ls = sorted(tree.get_leaf_nodes(), key=lambda x: x.details["box_ls"][0][0].tolist())
    temp = sorted(zip(box_ls, crop_ls), key=lambda x: x[0][0].tolist())
    sorted_crop_ls = [crop for _, crop in temp]
    sorted_box_ls = [box for box, _ in temp]

    for node, crop in zip(sorted_node_ls, sorted_crop_ls):
        node.var = crop

    tree.concat(beg_axis=beg_axis)
    whole = tree.var

    if paras["return_details"]:
        details = dict(
            whole=whole,
            crop_ls=sorted_crop_ls,
            box_ls=sorted_box_ls,
            beg_axis=beg_axis,
            computational_tree=tree,
        )
        return details
    else:
        return whole



import numpy as np


def convert_to_numpy(cfm, decimals=None):
    """
        将 confusion_matrices 从 tensor 转换为 numpy
    """
    for key, value in cfm.items():
        cfm[key] = value.cpu().numpy()
    # 由于 gpu 上的 tensor 通过 .cpu().numpy() 转换后会存在一定误差，因此需要进行进一步圆整
    if decimals is not None:
        cfm["thresholds"] = np.around(cfm["thresholds"], decimals)
    return cfm



def get_subsets(inputs):
    """
        获取所有的子集
    """
    sub_sets = [[]]
    for x in inputs:
        temp = []
        for item in sub_sets:
            temp.append(item + [x])
        sub_sets.extend(temp)
    return sub_sets


if __name__ == '__main__':
    print(get_subsets(range(20)))



import pytest
import numpy as np
import math
from kevin_toolbox.patches.for_test import check_consistency
from kevin_toolbox.math.number_theory import prime_factorization, get_primes


def test_prime_factorization():
    print("test number_theory.prime_factorization()")

    for _ in range(100):
        # 随机构建输入
        n = max(1, int(np.random.rand() * 1000))
        # 查找
        p_ls = prime_factorization(n=n)
        # 检查
        assert set(get_primes(n=n)).issuperset(set(p_ls))
        check_consistency(n, np.prod(p_ls))



from kevin_toolbox.math import utils


def merge_blocks(x, block_axis_num):
    """
        将最后 block_axis_num 个维度看做是 block，合并到此前 axis_num 个维度上。
            是 split_blocks() 的逆操作。

        参数：
            x:                  <np.array>
            block_axis_num:     <integer> 维度数量
    """

    # 校验参数
    _, function_table = utils.get_function_table_for_array_and_tensor(x)
    permute = function_table["permute"]
    assert isinstance(block_axis_num, (int,))
    assert x.ndim >= 2 * block_axis_num > 0

    offset = x.ndim - 2 * block_axis_num
    new_shape = list(x.shape[:offset])
    new_axis = list(range(offset))
    for i, (b, k) in enumerate(zip(x.shape[-2 * block_axis_num:-block_axis_num], x.shape[-block_axis_num:])):
        new_shape.append(b * k)
        new_axis.append(offset + i)
        new_axis.append(offset + i + block_axis_num)

    # x: [b, b0, b1, k0, k1] ==> [b, b0, k0, b1, k1] ==> [b, b0*k0, b1*k1]
    y = permute(x, new_axis).reshape(new_shape)
    return y



def scaling(**kwargs):
    """
        以给定的 zero_point 为原点，将 x 以 factor 为比例进行放大/缩小
            由于数值计算过程存在截断误差，本函数在同样的 factor,zero_point 配置下进行正向和逆向运算时，仅能保证 1e-2 之前的数值相同。

        必要参数：
            x:              <torch.tensor/np.array>
            factor:         <int/float>
            zero_point:     <int/float>
            reverse:        <boolean> 逆操作

        建议：
            - 对于需要保留更多小数点后精度的情况，建议在输入前先进行一定比例的放大。
    """

    # 默认参数
    paras = {
        # 必要参数
        "x": None,
        #
        "factor": 1,
        "zero_point": 0,
        "reverse": False,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    assert paras["x"] is not None
    x = paras["x"]
    for key in {"factor", "zero_point"}:
        assert isinstance(paras[key], (int, float,))
    assert paras["factor"] > 0
    factor, zero_point = paras["factor"], paras["zero_point"]

    if not paras["reverse"]:
        y = (x - zero_point) * factor + zero_point
    else:
        y = (x - zero_point) / factor + zero_point

    return y



class Executor:
    """
        执行器
            允许以静态方式描述并保存一个执行过程，
            然后在你需要的地方再进行调用。

        用法：
            1.定义执行过程
                executor = Executor( func = <function>,
                                        args = <list/tuple>,
                                        kwargs = <dict>,
                                        f_args = <list of functions>,
                                        f_kwargs = <dict of (key, function) pairs>)
                # 其中 func 是执行过程的主体函数
                # args 和 kwargs 是运行该函数时，要输入的参数
                # f_args 和 f_kwargs 的前缀 f_ 是 fixtures 固件的缩写
                #     固件内包含一系列的函数
                #     这些函数会在 func 执行前被首先执行，然后将得到的结果更新到参数中
                #     对于 f_args，执行后的结果会被 append 到 args 后面
                #     对于 f_kwargs，其中的 value 被执行后的结果将替换原来的 value，然后 update 到 kwargs 中
                # 使用 fixtures 的一大优势，就是可以让一些参数可以等到函数需要执行前再生成，从而节约资源
            2.调用执行过程
                executor.run()
                # 等效于 executor()
            3.修改执行过程
                当初始化的定义完成后，你还可以通过以下方式来修改执行过程：
                executor.set_paras( args=xx, ... )
                或者在使用过程中动态修改函数的输入参数：
                executor.run( input, reverse=True, xxx )  # 这里的参数 input, reverse 仅用作举例
                # 等效于：
                #       executor.set_paras( args=[input, ], kwargs=dict(reverse=True), ... )
                #       executor()

        注意！！
            对于 fixtures 中的函数，在定义函数时，函数体中如果涉及有外部的变量，
            则务必注意这些外部变量可能被修改，从而引起函数的行为发生不可预期的变化，
            例如：
                >> k=2
                >> y=lambda x:x**int(k)
                >> y(2)
                # 4
                >> k=4
                >> y(2)
                # 16
            解决方法：
                使用 Executor 来构造 fixtures 中的函数，同时使用 deepcopy 对参数进行隔离。
    """

    def __init__(self, **kwargs):
        f'{self.set_paras.__doc__}'

        # 默认参数
        self.paras = {
            # 必要参数
            "func": None,
            #
            "args": list(),
            "f_args": list(),
            "kwargs": dict(),
            "f_kwargs": dict(),
        }
        self.set_paras(**kwargs)

    def set_paras(self, **kwargs):
        """
            定义执行过程

            参数：
                func:                   <callable function> 函数
                args:                   <list/tuple> 参数
                f_args:                 <list of functions> “待解释”参数
                                            这里面的函数会在 func 执行前被首先执行，然后将得到的结果更新到参数 args 中。
                kwargs:                 <dict> 位置参数
                f_kwargs:               <dict of (key, function) pairs> “待解释”位置参数。
                                            这里面的函数会在 func 执行前被首先执行，然后将得到的结果更新到位置参数 kwargs 中。
        """
        # 默认参数
        paras = self.paras.copy()

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        # func
        assert paras["func"] is None or callable(paras["func"]), \
            f"func should be callable, but get a {type(func)}"
        # args
        assert isinstance(paras["args"], (list, tuple,)) and isinstance(paras["f_args"], (list, tuple,))
        for i, f in enumerate(paras["f_args"]):
            assert callable(f), \
                f"element {i} in f_args should be callable, but get a {type(f)}"
        # kwargs
        assert isinstance(paras["kwargs"], (dict,)) and isinstance(paras["f_kwargs"], (dict,))
        for k, v in paras["f_kwargs"].items():
            assert callable(v) and isinstance(k, (str,)), \
                f"item {key} in f_kwargs should be (str, callable) pairs, but get a ({type(key)}, {type(v)})"

        # update paras
        self.paras = paras

    def parse(self):
        assert callable(self.paras["func"]), \
            Exception(f"you should invoke set_paras() first, before calling parse()")

        # 获取函数
        func = self.paras["func"]

        # 获取参数
        args_, kwargs_ = [], dict()
        if "args" in self.paras:
            args_.extend(self.paras["args"])
        if "kwargs" in self.paras:
            kwargs_.update(self.paras["kwargs"])

        # evaluate the fixtures
        if "f_args" in self.paras:
            for f in self.paras["f_args"]:
                args_.append(f())
        if "f_kwargs" in self.paras:
            for k, v in self.paras["f_kwargs"].items():
                kwargs_[k] = v()

        return func, args_, kwargs_

    def run(self, *args, **kwargs):
        """
            调用执行过程
        """
        func, args_, kwargs_ = self.parse()
        assert callable(func), \
            Exception(f"you should invoke set_paras() first, before calling run()")

        # 根据当前输入动态更新参数
        if len(args) > 0:
            args_ = args
        kwargs_.update(kwargs)

        # 执行
        return func(*args_, **kwargs_)

    def __call__(self, *args, **kwargs):
        return self.run(*args, **kwargs)


if __name__ == '__main__':
    #
    executor = Executor(func=lambda x, y: print(x + y), args=[1], kwargs={"y": 3})
    print("executor")
    executor()
    #
    executor = Executor(func=lambda x, y: print(x, y), f_args=[lambda: 3], f_kwargs={"y": lambda: 4})
    print("executor using fixtures")
    executor()



import torch
from kevin_toolbox.env_info import version

"""
    参考 torch.concat()
        在 1.9 版本及其之前的 pytorch 中没有 torch.concat()，
        而只有 torch.cat()，
        两者用法与效果一致， concat 可以看做 cat 的别名
"""

if version.compare(torch.__version__, "<", "1.10", mode="short"):
    concat = torch.cat
else:
    concat = torch.concat



import pytest
import numpy as np
from kevin_toolbox.patches.for_test import check_consistency
from kevin_toolbox.computer_science.algorithm.combinatorial_optimization import get_subset_with_largest_product
from kevin_toolbox.computer_science.algorithm.for_seq import get_subsets


def test_get_subset_with_largest_product():
    print("test combinatorial_optimization.get_subset_with_largest_product()")

    for _ in range(500):
        # 随机构建输入的 ls、upper_bound
        end = int(np.random.randint(1, 100))
        ls = np.random.random(np.random.randint(1, 10)) * end
        upper_bound = np.random.random() * end * 2.3
        if np.random.randint(0, 2):
            ls = ls.astype(int)
            upper_bound = int(upper_bound)
        if min(ls) <= 0:
            ls = ls - min(ls) + 1e-8
        if upper_bound <= 0:
            upper_bound = 1e-8

        # 使用 get_subset_with_largest_product() 求解
        # print(ls, upper_bound)
        product, subset = get_subset_with_largest_product(
            ls=ls.tolist(), upper_bound=upper_bound,
        )
        # print(product, subset)

        # 通过暴力搜索获取真实答案
        best_product, best_subset_ls = None, []
        for temp in get_subsets(inputs=ls):
            temp_pd = np.prod(temp)
            if temp_pd > upper_bound:
                continue
            if best_product is None or best_product < temp_pd:
                best_product, best_subset_ls = temp_pd, [temp]
            elif temp_pd == best_product:
                best_subset_ls.append(temp)
        if best_product is None:
            best_subset_ls = [None]
        # print(best_product, best_subset_ls)

        # 检验
        check_consistency(product, best_product)
        if subset is not None:
            subset.sort()
            best_subset_ls = [sorted(i) for i in best_subset_ls]
        assert subset in best_subset_ls



import torch

# 计算设备（尽量使用gpu来加速计算）
device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')


def get_y_at_x(xs, ys, x, **kwargs):
    """
        对于 xs :=> ys 定义的离散函数，获取给定 x 下 y 的取值

        参数：
            xs：                 data
            ys：                 data or list of data
            mode：               插值模式
                                目前支持的模式：
                                    "pick_similar":   选取与 x 最相似的 xs 的取值，然后返回对应的 ys 中的值
            need_to_sort_xs:    是否需要先对 xs 进行排序
                                    在某些插值模式下，需要先对 xs 进行排序，然后才能进行计算
                                    默认为 False
    """
    mode = kwargs.get("mode", "pick_similar")
    assert mode in ["pick_similar", ]

    ys_ls = ys if isinstance(ys, (list, tuple,)) else [ys]
    assert len(xs) == max(map(len, ys_ls)) == min(map(len, ys_ls)) > 0

    xs = torch.tensor(xs, device=device, dtype=torch.float32) if not isinstance(xs, torch.Tensor) else xs
    for i, ys in enumerate(ys_ls):
        ys_ls[i] = torch.tensor(ys, device=device, dtype=torch.float32) if not isinstance(ys, torch.Tensor) else ys

    if kwargs.get("need_to_sort_xs", False):
        xs, sorted_indices = torch.sort(xs, dim=0)
        for i, ys in enumerate(ys_ls):
            ys_ls[i] = ys[sorted_indices.reshape(-1)]

    y_ls = []
    if mode == "pick_similar":
        gap = torch.abs(xs - x)
        index = torch.argmin(gap)
        for ys in ys_ls:
            y_ls.append(ys[index].cpu().numpy())
    assert len(y_ls) == len(ys_ls)

    res = y_ls[0] if len(y_ls) == 1 else y_ls
    return res


if __name__ == '__main__':
    import numpy as np

    xs_ = np.asarray([0.7, 0.5, 0.6, 0.4, 0.1, 0.3])
    ys_ = xs_ ** 2
    ys_2 = xs_ ** 3
    print(get_y_at_x(xs_, ys_, x=0.56))
    print(get_y_at_x(xs_, [ys_, ys_2], x=0.56))



from kevin_toolbox.computer_science.data_structure import Executor
from kevin_toolbox.machine_learning.dataset.face.verification.merge import merge


def get_executor_ls_by_block_along_diagonal(factory, i_0, i_1, j_0, j_1, chunk_step, need_to_generate,
                                            pick_triangle=True, include_diagonal=True):
    """
        通过调用 verification.Factory 中的 generate_by_block() 函数，
            来在（i_0, i_1, j_0, j_1）指定的子矩阵范围内，
            沿着 对角线，以 chunk_step 为间隔生成一系列的执行器 executor_ls，
            每个执行器在被 executor() 调用后都将返回一个数据集
        参数：
            factory:                verification.Factory 实例
            i_0, i_1, j_0, j_1:     子矩阵的范围
                                        （支持多种方式输入，具体参见 Factory.generate_by_block() 中的介绍）
            chunk_step:             chunk的长宽
            pre_res:                前置的数据集
                                        第一个chunk产生的数据集将与pre_res整合为一个dataset并返回
            need_to_generate:       需要生成的字段
                                        （参见 Factory.generate_by_block() 中的介绍）
            pick_triangle:          只取block的上三角
            include_diagonal:       是否包含对角线

        关于返回的数据集大小：
            当 pick_triangle=True 时，
            只取上三角
                将每两个 chunk 矩形的上三角，将两个上三角拼凑为一个dataset，有
                    dataset_size = chunk_step * (chunk_step + offset)
                    其中，当包含对角线上元素时，offset=1，不包含则为-1。
                当上三角的数量是奇数或者最后一个三角是残缺的话，
                    对于最后一个残缺的数据集，有
                    dataset_size <= chunk_step * (chunk_step + offset)
            当 pick_triangle=False 时，
            取整个block
                此时对于完整的矩形，有
                    dataset_size = chunk_step * chunk_step
                对于最后一个残缺的矩形，有
                    dataset_size <= chunk_step * chunk_step
    """
    i_len = i_1 - i_0 if isinstance(i_0, (int,)) else len(i_0)
    j_len = j_1 - j_0 if isinstance(j_0, (int,)) else len(j_0)
    assert 0 < i_len == j_len

    executor_ls, size_ls = [], []
    #
    count = 0
    while count < i_len:
        # 范围
        step = min(i_len - count, chunk_step)

        # 计算
        paras = dict(i_0=i_0 + count, i_1=i_0 + count + step,
                     j_0=j_0 + count, j_1=j_0 + count + step,
                     pick_triangle=pick_triangle,
                     include_diagonal=include_diagonal,
                     need_to_generate=need_to_generate)
        executor_ls.append(Executor(func=factory.generate_by_block,
                                    kwargs=paras))
        size_ls.append(factory.cal_size_of_block(**paras))
        #
        count += step

    # 合并
    for k in range(len(executor_ls) // 2):
        # 数据集
        u, v = executor_ls.pop(k), executor_ls.pop(k)
        executor_ls.insert(k, Executor(func=merge, f_args=[u, v]))
        # size
        u, v = size_ls.pop(k), size_ls.pop(k)
        size_ls.insert(k, u + v)

    return executor_ls, size_ls



import numpy as np
from kevin_toolbox.math.number_theory import cache
from kevin_toolbox.computer_science.algorithm.search import binary_search


def get_primes(n):
    """
        获取 小于等于 正整数n的所有素数
    """
    assert isinstance(n, (int, np.integer)) and n >= 1
    n += 1

    # 利用已有素数
    if n <= cache.primes[-1]:
        index = binary_search(cache.primes, n)
        return cache.primes[:index]

    # 求取素数
    for i in range(cache.primes[-1] + 1, n):
        b_prime = True
        for p, p_2 in zip(cache.primes[1:], cache.primes_square[1:]):
            if i % p == 0:
                b_prime = False
                break
            if p_2 > i:
                break
        if b_prime:
            cache.primes.append(i)
            cache.primes_square.append(i ** 2)
    return cache.primes[:]


if __name__ == '__main__':
    print(get_primes(n=271)[-1])
    print(get_primes(n=277)[-1])



import os

metadata_ls, content_ls, file_path_ls = [], [], []
from .data_0 import metadata, content

metadata_ls.append(metadata)
content_ls.append(content)
file_path_ls.append(os.path.join(os.path.abspath(os.path.dirname(__file__)), "data_0.txt"))



import numpy as np
import torch
from kevin_toolbox.data_flow.core.reader import Unified_Reader_Base, UReader

normalize = lambda x: x / np.sum(x ** 2, axis=1, keepdims=True) ** 0.5  # 归一化


class Dummy_Data_Factory:
    """
        用于生成人脸识别的伪数据

        背景介绍：
            在人脸识别任务中，我们假设每个人 human 都有一个表征 feature。
            而各个人在不同角度、环境、姿态下得到的人脸表征 feature 称为探针 probe。
            因此人脸识别实际上可以看做 probe feature 与 human feature 的匹配问题。

            有时候我们并不知道所有 human 的 feature，而只知道一部分。
            我们就将已知的 human feature 构成一个人脸库 gallery。
            判断 probe 是否在 gallery 中也可以看做一个二分类任务。

        实例化时输入的变量（特征库）：
            humans:                本征特征。
                                        shape [human_nums, dims_of_feature]
            gallery:               特征库中哪些是已知的。
                                        shape [human_nums, 1]
                                        已知的（亦即是库内人脸将被标记为1，反之为0）
        以上均可根据以下参数来随机生成：
            human_nums:             不同 human 的数量
            gallery_nums:           人脸库中 human 的数量
            dims_of_feature:        特征的维度

        构造数据集时输入的参数：
            probe_nums:             探针样本的数量
            noise_significance:     生成探针样本时，添加噪声的比例

        构造数据集时生成的变量（按照 probe_id 进行排序，一一对应）：
            probes（features）:     人脸样本。
                                        从 humans 中独立随机抽取 probe_nums 次，将抽得的特征再加以噪声糅合而成
                                        shape [probe_nums, dims_of_feature]
            clusters:               probe 所属的 human_id
                                        shape [probe_nums, 1]
            in_gallery:             probe 是否在 gallery 中
                                        shape [probe_nums, 1]

        用法：
            1. __init__() 在初始化时，输入相关变量
                factory = Dummy_Data_Factory( human_nums=xxx, gallery_nums=xxx )
            2. 通过 generate 函数来生成数据集
                dataset = factory.generate( nums=xxx, noise_significance=xxx,  need_to_generate=xxx)
                # 返回的数据集是一个字典
                # dataset := {"features": xxx, "clusters": xxx, "in_gallery": xxx}
                # 其中具体包含哪些字段，可以通过添加 need_to_generate 参数来指定
    """

    def __init__(self, *args, **kwargs):
        """
            构建人脸特征库
            参数:
                humans:             本征特征。
                                        shape [human_nums, dims_of_feature]
                                        需要被 Unified_Reader_Base 包裹
                                        不指定时，根据 human_nums 随机生成
                human_nums:         不同 human 的数量
                dims_of_feature:    特征的维度

                gallery:            human 是否在 gallery 中
                                        shape [human_nums, 1]
                                        不指定时，根据 gallery_nums 随机生成
                gallery_nums:       人脸库中 human 的数量
                                        默认等于 human_nums
        """

        # 默认参数
        paras = {
            # human
            "humans": None,
            "human_nums": None,
            "dims_of_feature": 256,
            # gallery
            "gallery": None,
            "gallery_nums": None,
        }

        # 获取参数
        paras.update(kwargs)

        # 计算设备（尽量使用gpu来加速计算）
        self.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

        # 校验参数
        # human [human_nums, dims_of_feature]
        if paras["humans"] is None:
            assert isinstance(paras["human_nums"], (int,))
            if isinstance(paras["dims_of_feature"], (int,)):
                paras["dims_of_feature"] = [paras["dims_of_feature"]]
            paras["dims_of_feature"] = list(paras["dims_of_feature"])
            # 构建人脸特征库
            paras["humans"] = UReader(var=normalize(
                np.random.uniform(0, 1, [paras["human_nums"]] + paras["dims_of_feature"]).astype(np.float32)))
        assert isinstance(paras["humans"], (Unified_Reader_Base,))

        # gallery [human_nums, 1]
        if paras["gallery"] is None:
            assert isinstance(paras["gallery_nums"], (int,))
            rand_index = np.random.choice(np.arange(paras["human_nums"]), size=paras["gallery_nums"], replace=False)
            gallery = np.zeros(shape=[paras["human_nums"], 1])
            gallery[tuple(rand_index),] = 1
            paras["gallery"] = UReader(var=gallery)
        assert isinstance(paras["gallery"], (Unified_Reader_Base,))

        self.paras = paras

    def generate(self, nums, **kwargs):
        """
            生成人脸数据集

            参数：
                nums:               样本的数量
                noise_significance:     生成探针样本时，添加噪声的比例
                need_to_generate:   指定数据集中需要生成的字段
                                        目前支持的字段有： {"features", "clusters", "in_gallery"}
        """

        __support_to_generate = {"features", "clusters", "in_gallery"}
        need_to_generate = kwargs.get("need_to_generate", __support_to_generate)
        assert isinstance(need_to_generate, (set,)) and need_to_generate.issubset(__support_to_generate)

        noise_significance = kwargs.get("noise_significance", 0.2)
        assert 0 <= noise_significance <= 1

        res = dict()

        # 计算 clusters
        clusters = np.random.randint(0, len(self.paras["humans"]), [nums, 1])

        if "clusters" in need_to_generate:
            res["clusters"] = clusters  # shape: [ nums, 1 ]

        # 计算 features
        if "features" in need_to_generate:
            true_features = self.paras["humans"].read(list(clusters.reshape(-1)))
            noise = np.random.uniform(0, 1, true_features.shape)
            features = noise_significance * noise + (1 - noise_significance) * true_features  # 为probe中的特征添加噪声，使其偏离真实值
            features = normalize(features)  # 归一化
            res["features"] = features  # shape: [ nums, dims_of_feature ]

        # 计算 in_gallery
        if "in_gallery" in need_to_generate:
            in_gallery = self.paras["gallery"].read(list(clusters.reshape(-1)))
            res["in_gallery"] = in_gallery  # shape: [ nums, 1 ]

        return res



import torch
import numpy as np


def _init_var(like=None, data_format=None):
    if like is not None:
        if torch.is_tensor(like):
            var = torch.zeros_like(like)
        elif isinstance(like, (np.ndarray,)):
            var = np.zeros_like(like)
        elif isinstance(like, (int, float, np.number,)):
            var = 0.0
        else:
            raise ValueError("paras 'like' should be np.ndarray, torch.tensor or int/float")
    elif data_format is not None:
        assert isinstance(data_format, (dict,)) and "type_" in data_format and "shape" in data_format
        k_s = copy.deepcopy(data_format)
        k_s.pop("type_")
        k_s.pop("shape")
        if data_format["type_"] == "torch":
            var = torch.zeros(size=data_format["shape"], **k_s)
        else:
            var = np.zeros(shape=data_format["shape"], **k_s)
    else:
        var = None

    return var



import numpy as np


def cal_iou_between_box(**kwargs):
    """
        计算 box_0 和 box_1 之间的交并比 iou

        参数：
            box_0:          <np.array>
                                shape [2, dimensions]，各个维度的意义为：
                                    2：          box的两个轴对称点
                                    dimensions： 坐标的维度
            box_1:          <np.array>
                                与 box_0 类似。
            return_details: <boolean> 是否以详细信息的形式返回结果
                                默认为 False，此时返回：
                                    iou <float>
                                当设置为 True，将返回一个 dict：
                                    details = dict(
                                        iou=<float>,
                                        intersection=dict(area=<float>, box=<np.array>,),  # area 表示面积
                                        union=dict(area=<float>),
                                        box_0=dict(area=<float>, box=<np.array>,),
                                        box_1=dict(area=<float>, box=<np.array>,),
                                    )
    """
    # 默认参数
    paras = {
        # 必要参数
        "box_0": None,
        "box_1": None,
        "return_details": False,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    for key in ("box_0", "box_1"):
        paras[key] = np.asarray(paras[key])
        assert paras[key].ndim == 2 and paras[key].shape[0] == 2
    assert paras["box_0"].shape[1] == paras["box_1"].shape[1]
    box_0, box_1 = paras["box_0"], paras["box_1"]
    for box in (box_0, box_1):
        box.sort(axis=0)

    # intersection
    beg = np.maximum(box_0[0], box_1[0])
    end = np.minimum(box_0[1], box_1[1])
    edge_lens = np.maximum(end - beg, 0)
    intersection = edge_lens.prod()

    # union
    union = 0
    for box in (box_0, box_1):
        union += (box[1] - box[0]).prod()
    union -= intersection

    # iou
    if union > 0:
        iou = intersection / union
    else:
        iou = 0

    if paras["return_details"]:
        details = dict(
            iou=iou,
            intersection=dict(area=intersection, box=np.asarray([beg, end])),
            union=dict(area=union),
            box_0=dict(area=(box_0[1] - box_0[0]).prod(), box=box_0),
            box_1=dict(area=(box_1[1] - box_1[0]).prod(), box=box_1),
        )
        return details
    else:
        return iou



from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d.proj3d import proj_transform
import numpy as np


class Arrow3D(FancyArrowPatch):
    """
        适用于3d图的 360度 无死角箭头
        来自：https://gist.github.com/WetHat/1d6cd0f7309535311a539b42cccca89c

        使用方法：
            arrow = Arrow3D(*beg.tolist(), *(end - beg).tolist(),
                    arrowstyle="-|>", mutation_scale=15,
                    color="blue", alpha=0.5)
            ax3d.add_artist(arrow)
    """

    def __init__(self, x, y, z, dx, dy, dz, *args, **kwargs):
        super().__init__((0, 0), (0, 0), *args, **kwargs)
        self._xyz = (x, y, z)
        self._dxdydz = (dx, dy, dz)

    def draw(self, renderer):
        x1, y1, z1 = self._xyz
        dx, dy, dz = self._dxdydz
        x2, y2, z2 = (x1 + dx, y1 + dy, z1 + dz)

        xs, ys, zs = proj_transform((x1, x2), (y1, y2), (z1, z2), self.axes.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        super().draw(renderer)

    def do_3d_projection(self, renderer=None):
        x1, y1, z1 = self._xyz
        dx, dy, dz = self._dxdydz
        x2, y2, z2 = (x1 + dx, y1 + dy, z1 + dz)

        xs, ys, zs = proj_transform((x1, x2), (y1, y2), (z1, z2), self.axes.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))

        return np.min(zs)





def _recursive(indices, axis_beg, axis_end, step, shape):
    if axis_beg == axis_end:  # 到达最后一维
        for i in range(shape[axis_beg]):
            indices[axis_beg] = i
            yield indices
    else:
        for i in range(shape[axis_beg]):
            indices[axis_beg] = i
            yield from _recursive(indices, axis_beg + step, axis_end, step, shape)


def normal_indices_generator(shape, order="C"):
    """
        迭代生成遍历 shape 所需要的所有的坐标 indices
            indices 格式的具体定义参见 coordinates.convert()
        
        参数：
            shape:          <list/tuple> 要遍历的形状
            order:          <str> 遍历的模式
                                两种取值：
                                    "C":    row first，从前往后，首先遍历第一个维度
                                    "F":    column first，从后往前，首先遍历最后一个维度
    """
    assert isinstance(shape, (list, tuple,)) and len(shape) > 0
    assert order in ["C", "F"]

    if order == "C":
        axis_beg, axis_end = 0, len(shape) - 1
        step = 1
    else:
        axis_end, axis_beg = 0, len(shape) - 1
        step = -1

    for indices in _recursive(indices=[0] * len(shape), axis_beg=axis_beg, axis_end=axis_end, step=step, shape=shape):
        yield indices[:]



def zero_one_knapsack_problem(**kwargs):
    """
        使用动态规划求 01 背包
            支持 weights 和 values 是负数的情况

        参数:
            weights:                <list> 可选 item 的“体积”
            values:                 <list> 对应 item 的价值
                                        注意：weights 和 values 中也可以包含负数
            upper_bound:            <int/float> 背包的“容量”上限
                                        注意：upper_bound 也可以是负数
        返回:
            v, idx_ls
            背包可以容纳的最大价值，对应子集的下标序列
                当解不存在时候，返回 None, None
    """
    # 默认参数
    paras = {
        "weights": None,
        "values": None,
        "upper_bound": None,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    assert paras["weights"] is not None
    assert paras["upper_bound"] is not None
    if paras["values"] is None:
        paras["values"] = paras["weights"]
    assert len(paras["weights"]) == len(paras["values"]) > 0

    return _recursion(weights=paras["weights"], values=paras["values"], upper_bound=paras["upper_bound"],
                      end_idx=len(paras["weights"]) - 1, memory=dict())


def _recursion(weights, values, upper_bound, end_idx, memory):
    if end_idx < 0:
        return (None, None) if upper_bound < 0 else (0, [])
    if (upper_bound, end_idx) in memory:
        return memory[(upper_bound, end_idx)]

    v_0, idx_ls_0 = _recursion(weights=weights, values=values, upper_bound=upper_bound, end_idx=end_idx - 1,
                               memory=memory)
    v_1, idx_ls_1 = _recursion(weights=weights, values=values, upper_bound=upper_bound - weights[end_idx],
                               end_idx=end_idx - 1, memory=memory)
    if v_1 is not None:
        v_1 += values[end_idx]
        idx_ls_1 = idx_ls_1 + [end_idx]

    if v_0 is None:
        res = (v_1, idx_ls_1)
    elif v_1 is None:
        res = (v_0, idx_ls_0[:])
    else:
        res = (v_0, idx_ls_0[:]) if v_0 >= v_1 else (v_1, idx_ls_1)

    memory[(upper_bound, end_idx)] = res

    return res


if __name__ == '__main__':
    # print(zero_one_knapsack_problem(weights=[3, -1, 6, 5, 4], values=[1, 1, 1, 1, 1], upper_bound=7))
    # print(zero_one_knapsack_problem(weights=[-3, -1, -6, -5, -4], upper_bound=7))
    # print(zero_one_knapsack_problem(weights=[0, -3, 3], values=[0, 0, 5], upper_bound=2))
    # print(zero_one_knapsack_problem(weights=[3., 3., 3.], upper_bound=6.338211323106079))
    print(zero_one_knapsack_problem(weights=[0, -5, 3, 0, -1, 3], values=[0, -3, -4, -1, -2, 2], upper_bound=-7))
    print(zero_one_knapsack_problem(weights=[1.2, 0.4, 2, 2, 2, 2, 2, 2, 2, 2], upper_bound=5))



from kevin_toolbox.math.dimension import transpose


def transpose_inside_axis(**kwargs):
    """
        将变量 x 的第 axis 个轴内的各个维度，按照 index_ls 的顺序进行重排/转置

        参数：
            x:          <np.array/torch.tensor>
            axis:       <integer>
                            默认为 -1
            index_ls:   <list> 格式具体参考 dimension.coordinates
            reverse:    <boolean>
                            默认为 False
                            当设置为 True 时进行转置的逆

        建议：
            - 首先将 x 中需要重排的 axis 转置到最后能够加速本函数的执行。
    """
    # 默认参数
    paras = {
        # 必要参数
        "x": None,
        "index_ls": None,
        #
        "axis": -1,
        "reverse": False,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    assert paras["x"] is not None
    x = paras["x"]
    #
    if paras["axis"] < 0:
        paras["axis"] += x.ndim
    assert isinstance(paras["axis"], (int,)) and 0 <= paras["axis"] < x.ndim
    axis = paras["axis"]
    #
    assert len(paras["index_ls"]) == x.shape[axis], \
        f'{len(paras["index_ls"])} == {x.shape[axis]}？'
    index_ls = paras["index_ls"]
    if paras["reverse"]:
        index_ls = transpose.get_inverse_index_ls(index_ls=index_ls)

    slices = tuple([slice(None, None)] * axis + [index_ls])
    y = x[slices]
    return y



import torch
from kevin_toolbox.env_info import version

"""
    tile(x, multiples)
    
    将 input 的各个维度按照 multiples 中的数速进行复制（占用内存）
        由于在1.7版本及其之前的 pytorch 没有 torch.tile，
        考虑到兼容性，本函数在pytorch版本过低时自动使用 x.repeat() 来实现 tile 函数

    参数：
        x:                  input tensor
        multiples:          要复制的维度
"""

if version.compare(torch.__version__, "<", "1.8", mode="short"):
    def tile(x, multiples):
        assert isinstance(multiples, (list, tuple,))
        multiples = list(multiples) + [1] * (x.ndim - len(multiples))
        return x.repeat(*multiples)
else:
    def tile(x, multiples):
        assert isinstance(multiples, (list, tuple,))
        return torch.tile(x, multiples)



from kevin_toolbox.math import utils


def split_blocks(x, block_shape):
    """
        将最后部分的维度 axis 按照 block_shape 分割成 blocks 的组成
            例如，对于 x=[5, 6, 6]，在 block_shape=[3, 2] 的情况下将得到 y=[5, 2, 3, 3, 2]

        参数：
            x:                  <np.array>
            block_shape:        <list/tuple> block 的形状
    """

    # 校验参数
    _, function_table = utils.get_function_table_for_array_and_tensor(x)
    permute = function_table["permute"]
    assert isinstance(block_shape, (tuple, list,))
    assert x.ndim >= len(block_shape)

    new_shape = list(x.shape[:x.ndim - len(block_shape)])
    axis_ids_for_group, axis_ids_for_block = [], []
    for i, j in zip(x.shape[-len(block_shape):], block_shape):
        assert i >= j and i % j == 0
        new_shape.append(i // j)
        axis_ids_for_group.append(len(new_shape) - 1)
        new_shape.append(j)
        axis_ids_for_block.append(len(new_shape) - 1)
    new_axis = list(range(x.ndim - len(block_shape))) + axis_ids_for_group + axis_ids_for_block

    # x: [b, w, h] ==> [b, w/k0, k0, h/k1, k1] ==> [b, w/k0, h/k1, k0, k1]
    y = permute(x.reshape(new_shape), new_axis)
    return y



from .my_iterator_base import My_Iterator_Base


class My_Iterator(My_Iterator_Base):
    """
        My_Iterator_Base 的简单实现
    """

    def __init__(self, array):
        assert isinstance(array, (list, tuple,))
        self.array = array
        self.beg, self.end, self.offset = 0, len(self.array), 0

    # ------------------------------------ read ------------------------------------ #

    def read(self, index):
        """
            从 array 中读取对应位置的元素
                根据实际需要来修改，比如添加类型检查、转换等
        """
        return self.array[index]

    # ------------------------------------ iterator ------------------------------------ #

    def __iter__(self):
        return self

    def __next__(self):
        index = self.offset + self.beg
        if index >= self.end:
            raise StopIteration

        v = self.read(index)
        self.offset += 1
        return v

    def set_range(self, beg, end):
        assert 0 <= beg <= end <= len(self.array)
        self.beg = beg
        self.end = end
        self.offset = 0

    def pass_by(self, num):
        self.offset += num

    # ------------------------------------ sequence ------------------------------------ #

    def __getitem__(self, index):
        index = self.__round_by_range(self.beg, self.end, index)
        v = self.read(index)
        return v

    def __len__(self):
        return self.end - self.beg

    # ------------------------------------ else ------------------------------------ #

    def __round_by_range(self, beg, end, offset):
        return super().round_by_range(beg, end, offset)



RAW_SETTINGS = dict(
    default=dict(
        read=lambda x: x,
        write=lambda x: f"{x}",
    ),
    int=dict(
        read=lambda x: int(eval(x)),
        write=lambda x: f"{int(x) if x is not None else None}",
    ),
    float=dict(
        read=lambda x: float(x.replace("None", "nan")),
    ),
    str=dict(
        read=lambda x: str(x),
    ),
    list=dict(
        read=lambda x: eval(deal_nan(x, None)),
    ),
)


def deal_nan(x, replace_with=None):
    x = x.replace("nan", str(replace_with))
    return x


SETTINGS = dict(
    read=dict(),
    write=dict()
)
for converter_name, item in RAW_SETTINGS.items():
    for k, v in item.items():
        SETTINGS[k][converter_name] = v


class Converter:
    def __init__(self, **kwargs):
        """
            获取具体的 converter
                converter is a dictionary-like data structure consisting of <string>:<func> pairs，
                用于根据指定数据类型选取适当的函数来处理输入数据

            参数：
                mode：           <string> 转换模式
                                    可选值：
                                        "read":     适用于 Kevin_Notation_Reader
                                        "write":    适用于 Kevin_Notation_Writer
        """
        # 默认参数
        paras = {
            # 必要参数
            "mode": None,
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        assert paras["mode"] in ["read", "write"]

        self.paras = paras

    def get(self, key, **kwargs):
        """
            获取具体的 converter
                converter is a dictionary-like data structure consisting of <string>:<func> pairs，
                用于根据指定数据类型选取适当的函数来处理输入数据

            参数：
                key:            <string> 键
                use_default:    <boolean> 当无法查找到 key 对应的 converter 时，是否返回默认的 converter
                                    默认为 True
                                    当设置为 False 且遇到不存在的 key 时，将引起 KeyError 错误
                mode：           <string> 另外设定模式，而不使用初始化时默认设定的模式
        """

        # 默认参数
        paras = {
            "use_default": True,
            "mode": self.paras["mode"],
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        assert isinstance(key, (str,))
        assert paras["mode"] in ["read", "write"]

        converter = SETTINGS[paras["mode"]].get(key, None)
        if converter is None:
            if paras["use_default"]:
                converter = SETTINGS[paras["mode"]].get("default")
            else:
                raise KeyError(f"the most recent CONVERTER does not support type {key}")

        return converter

    def __getitem__(self, key):
        """
            通过 self[] 调用 get()
        """
        return self.get(key)

    def keys(self):
        return SETTINGS[self.paras["mode"]].keys()

    def values(self):
        return SETTINGS[self.paras["mode"]].values()

    def len(self):
        return len(SETTINGS[self.paras["mode"]])

    def items(self):
        return SETTINGS[self.paras["mode"]].items()


CONVERTER_FOR_READER = Converter(mode="read")
CONVERTER_FOR_WRITER = Converter(mode="write")

if __name__ == '__main__':
    print(SETTINGS)

    converter_ = Converter(mode="read")

    print(converter_.get("float"))
    print(converter_["f"])
    print(converter_.keys())



import re
import inspect
import pkgutil
import weakref
from kevin_toolbox.computer_science.algorithm.for_nested_dict_list import get_value_by_name, set_value_by_name, \
    traverse, count_leaf_node_nums


class Registry:
    """
        注册器
            具有以下功能
            - 管理成员，包括添加 add()、获取 get() pop() 成员等
            - 支持通过装饰器 register() 来添加成员
            - 支持通过 collect_from() 搜索指定的路径，当该路径下的模块被 register() 装饰器包裹时，将自动导入（用于解决python中的模块是惰性的问题）

        使用方法：
            以目录结构：
                xx/——— modules_dir
                 |       |——— a.py
                 |       |___ ...
                 |——— test.py
            为例。
            我们的目标是在 test.py 中，构建一个能够自动加载 modules_dir 中待注册成员的 Registry 实例。
            具体步骤如下：
                1. 在文件 test.py 中：
                    # 创建注册器实例
                    DB = Registry(uid="DB")
                    # 设置搜索路径
                    DB.collect_from(path_ls=["xx/modules_dir"])

                2. 在 modules_dir 下需要加载的成员中，以 a.py 为例：
                    # 导入注册器实例
                    from xx.test import DB
                    # 使用 DB.register() 装饰器注册
                    @DB.register(name=":my_module")
                    class A:
                        ...
                    # 使用 add() 函数注册
                    DB.add(obj=233, name=":var:c")

                3. 在文件 test.py 中：
                    # 获取已注册成员
                    module = DB.get(name=":my_module")

                4. 如果需要在其他文件中获取已有注册器实例，除了用 import 之外，还可以直接用 uid
                    # 获取指定 uid 下的实例
                    temp = Registry(uid="DB")
    """
    name = "registry made by kevin"
    __instances = weakref.WeakValueDictionary()  # {uid: instance_i, ...} 用于保存创建的实例
    __counter = 0

    def __new__(cls, *args, **kwargs):
        """
            __new__函数返回的实例，将作为self参数被传入到__init__函数。
                如果__new__函数返回一个已经存在的实例（不论是哪个类的），__init__还是会被调用的，所以要特别注意__init__中对变量的赋值。
        """
        uid = kwargs.get("uid", cls.__counter)

        # 获取实例
        if uid in cls.__instances:
            # 返回已存在的实例
            self = cls.__instances[uid]
        else:
            # 传入 __init__ 中新建一个实例
            self = super().__new__(cls)
        return self

    def __init__(self, *args, **kwargs):
        """
            参数：
                uid:                <hashable> 实例的唯一标识符
                                        默认置空，此时将新建一个实例，并将此时已有实例的数量作为 uid
                                        当为非空，则根据 uid 到已有实例中寻找相同的实例，并返回（使用该特性可以构造单例模式）
        """
        try:
            getattr(self, "uid")
        except:
            pass
        else:
            return  # 如果是从 __new__ 中获取的已有的实例则不重复进行参数赋值
        self.database = dict()
        #
        self.uid = kwargs.get("uid", Registry.__counter)
        self.inactivate_b_force_of_add = False
        # 记录到 __instances 中
        Registry.__instances[self.uid] = self
        Registry.__counter += 1

    def add(self, obj, name=None, b_force=False):
        """
            注册

            参数：
                obj：            待注册成员
                                    可以是函数、类或者callable的实例
                                    也可以是各种 int、float、str变量
                                    总之一切对象皆可
                name：           <str> 成员名称
                                    默认为 None，此时将从被注册对象 obj 的属性中尝试推断出其名称。
                                        若 obj 中有 name 或者 __name__ 属性（优先选择name），则推断出的名称是 f':{obj.name}'；
                                            进一步若有 version 属性，则为 f':{obj.name}:{obj.version}'
                                        否则报错。
                                        比如下面的类：
                                            class A:
                                                version="1.0"
                                        的默认注册名称将是 ":A:1.0"
                                    对于 int、str 和其他没有 name 或者 __name__ 属性的变量则必须要手动指定 name 参数。
                        需要注意的是，成员的名称确定了其在注册器内部 database 中的位置，名称的解释方式参考 get_value_by_name() 中的介绍。
                        因此不同的名称可能指向了同一个位置。
                b_force：          <boolean> 是否强制注册
                                    默认为 False，此时当 name 指向的位置上已经有成员时，将不进行覆盖而直接报错
                                    当设置为 True，将会强制覆盖
        """
        # 检验参数
        if name is None:
            name = getattr(obj, "name", getattr(obj, "__name__", None))
            if name is not None:
                name = f':{name}'
                version = getattr(obj, "version", None)
                if version is not None:
                    name += f':{version}'
        assert isinstance(name, (str,))

        # 尝试注册
        temp = self.database.copy()
        set_value_by_name(var=temp, name=name, value=obj, b_force=True)
        # check
        if not self.inactivate_b_force_of_add and not b_force:
            inc_node_nums = count_leaf_node_nums(var=obj) if isinstance(obj, (list, dict)) else 1  # 增加的节点数量
            assert count_leaf_node_nums(var=temp) == count_leaf_node_nums(var=self.database) + inc_node_nums, \
                f'registration failed, it may be a conflict with an existing member'
        self.database = temp

    def get(self, name, **kwargs):
        """
            获取

            参数：
                name：           <str> 成员名称
                default:          默认值
                                    找不到时，若无默认值则报错，否则将返回默认值
        """
        try:
            return get_value_by_name(var=self.database, name=name)
        except:
            if "default" in kwargs:
                return kwargs["default"]
            else:
                raise AssertionError(f'element {name} not exists')

    def pop(self, name, **kwargs):
        """
            弹出

            参数：
                name：           <str> 成员名称
                                    名称一定要准确，不能含有模糊的 | 字符
                default:          默认值
                                    找不到时，若无默认值则报错，否则将返回默认值
        """
        assert "|" not in name, \
            f'to pop up a specific member needs to give the exact name, ' \
            f'does not support the "|" character with automatic deduction, invalid name: {name}, ' \
            f'what you probably want to pop is {name.replace("|", ":")}'
        res, b_found = None, False

        def func(_, idx, value):
            nonlocal res, b_found
            if res is None and idx == name:
                res, b_found = value, True
                return True
            else:
                return False

        traverse(var=self.database, match_cond=func, action_mode="remove", b_use_name_as_idx=True)
        if b_found:
            return res
        else:
            if "default" in kwargs:
                return kwargs["default"]
            else:
                raise AssertionError(f'element {name} not exists')

    def clear(self):
        self.database.clear()

    # -------------------- 装饰器 --------------------- #

    def register(self, name=None, b_force=False):
        """
            用于注册成员的装饰器
                成员可以是函数、类或者callable的实例

            参数：
                name：           <str> 成员名称
                                    默认为 None
                b_force：        <boolean> 是否强制注册
                                    默认为 False
                                （以上参数具体参考 add() 函数介绍）
        """

        def wrapper(obj):
            nonlocal self, name, b_force
            self.add(obj, name=name, b_force=b_force)
            return obj

        return wrapper

    # -------------------- 其他 --------------------- #

    def collect_from(self, path_ls):
        """
            遍历 path_ls 下的所有模块，并自动导入其中被 register() 装饰器包裹的部分
                注意，自动导入时使用的 add() 方法中的 b_force 将强制设置为 True
                这意味着不再检查可能冲突的添加项
        """
        temp = None
        for loader, module_name, is_pkg in pkgutil.walk_packages(path_ls):
            module = loader.find_module(module_name).load_module(module_name)
            if temp is None:
                for name, obj in inspect.getmembers(module):
                    if getattr(obj, "name", None) == Registry.name and getattr(obj, "uid") == self.uid:
                        temp = obj
                        temp.inactivate_b_force_of_add = True
                        break
        if temp is not None:
            self.database = temp.database


UNIFIED_REGISTRY = Registry(uid="UNIFIED_REGISTRY")



from .node import Node
from .sender import Sender
from .getter import Getter


class Graph:
    def __init__(self):
        self.getter_db = dict()
        self.sender_db = dict()
        self.node_db = dict()

    def add_node(self, exist_ok=False, uid=None, **kwargs):
        assert isinstance(uid, (str, type(None),)), \
            TypeError(f"uid should be string or None, but get a {type(uid)}!")

        if uid is None or uid not in self.node_db:
            # uid不存在，新建节点
            node = Node(uid=uid, **kwargs)
        else:
            # uid已存在
            if not exist_ok:
                raise Exception(f"Node {kwargs['uid']} already exists!")
            node = self.node_db[uid]
        return node


from .generate_integrated_api import generate_integrated_api
from .generate_shuffled_index_ls import generate_shuffled_index_ls
from .generate_z_pattern_indices_ls import generate_z_pattern_indices_ls
from .normal_indices_generator import normal_indices_generator



from kevin_toolbox.data_flow.file import kevin_notation


def write(metadata, content, file_path):
    """
        写入整个文件的快捷接口
    """
    with kevin_notation.Writer(file_path=file_path, mode="w", sep=metadata.get("sep", "\t")) as writer:
        writer.write_metadata(metadata=metadata)
        if isinstance(content, (dict,)):
            writer.write_contents(column_dict=content)  # 字典方式写入
        else:
            writer.write_contents(row_ls=content)  # 列表方式写入



import numpy as np
import torch


def convert_ndarray_to_list(x):
    """
        尝试将 numpy 或者 torch 数组转换为 list
    """

    if type(x) is np.ndarray:
        return x.tolist()
    elif torch.is_tensor(x):
        return x.detach().cpu().numpy().tolist()
    else:
        return x


if __name__ == '__main__':
    print(convert_ndarray_to_list(torch.ones(3, 4).cuda()))



import pytest
import numpy as np
import math
from kevin_toolbox.math.number_theory import get_primes
from kevin_toolbox.patches.for_test import check_consistency


def test_get_primes():
    print("test number_theory.get_primes()")

    for _ in range(100):
        # 随机构建输入
        n = max(1, int(np.random.rand() * 1000))
        # 查找
        p_ls = get_primes(n=n)
        # 标准答案
        p_ls_1 = []
        for p in range(1, n + 1):
            b_is_prime = True
            for i in range(2, min(math.ceil(p ** 0.5) + 1, p)):
                if p % i == 0:
                    b_is_prime = False
                    break
            if b_is_prime:
                p_ls_1.append(p)
        # 检查
        check_consistency(sorted(p_ls), sorted(p_ls_1))



from kevin_toolbox.computer_science.algorithm.for_nested_dict_list import get_value_by_name


def set_value_by_name(var, name, value, b_force=False):
    """
        通过解释名字得到取值方式，然后到 var 中将对应部分的值修改为 value。

        参数：
            var:            任意支持索引赋值的变量
            name:           <string> 取值方式
                                由多组 "<解释方式><键>" 组成。
                                解释方式支持以下几种:
                                    "@"     使用 eval() 读取键
                                    ":"     使用 str() 读取键
                                    "|"     依次尝试 str() 和 eval() 两种方式
                                示例:
                                    "@100"      表示指向 var[eval("100")]
                                    ":epoch"    表示指向 var["epoch"]
                                    "|1+1"    表示首先尝试指向 var["1+1"]，若不成功则尝试指向 var[eval("1+1")]
                                假设 var=dict(acc=[0.66,0.78,0.99])，如果你想将 var["acc"][1] 设置为 100，那么可以将 name 写成：
                                    ":acc@1" 或者 "|acc|1" 等。
                                注意，在 name 的开头也可以添加任意非解释方式的字符，本函数将直接忽略它们，比如下面的:
                                    "var:acc@1" 和 "xxxx|acc|1" 也能正常写入。
            value:          待赋给的值
            b_force:        <boolean> 当无法设置时，是否尝试创建或者修改节点
                                默认为 False，
                                当设置为 True，可能会对 var 的结构产生不可逆的改变，请谨慎使用。新创建或者修改的节点的类型是 dict。
    """
    assert isinstance(name, (str,))
    key = name
    for char in ":@|":
        key = key.rsplit(char, 1)[-1]
    assert len(key) < len(name), f'invalid name: {name}'

    try:
        item = get_value_by_name(var=var, name=name[:-1 - len(key)])

        flag = name[-1 - len(key)]
        key_t = key
        if flag == "@":
            key_t = eval(key)
        elif flag == "|":
            try:
                _ = item[key]
            except:
                key_t = eval(key)

        item[key_t] = value
    except:
        if not b_force:
            raise ValueError(f'The location pointed to by name {name} does not exist in var')
        else:
            var = set_value_by_name(var=var, name=name[:-1 - len(key)], value={key: value}, b_force=b_force)

    return var



def flatten_list(ls, depth=None):
    """
        将列表展平
            比如对于输入 ls=[[1, 2, (3, 4)], [(5,)], 6], depth=None
            展平后得到的输出为 [1, 2, 3, 4, 5, 6]

        参数:
            ls:         <list/tuple> 要展平的列表
            depth:      <int> 展开到第几层
                            0 表示当前层，亦即不作展开
                            默认为 None 表示展开到最深的那层
    """
    assert isinstance(ls, (list, tuple,))
    assert depth is None or isinstance(depth, (int,))
    return _recursion(ls=ls, depth=depth)


def _recursion(ls, depth):
    if depth is not None and depth <= 0:
        res = ls
    else:
        res = []
        for i in ls:
            if isinstance(i, (list, tuple,)):
                res.extend(_recursion(i, depth=depth if depth is None else depth - 1))
            else:
                res.append(i)
    return res


if __name__ == '__main__':
    a = [[1, 2, (3, 4)], [(5,)], 6]
    print(flatten_list(ls=a, depth=None))



import numpy as np


def normalize(v, ord, axis=-1):
    """
        对变量沿着轴 axis 进行归一化

        参数：
            v：
            ord：            表示范数的类型。可以设置为1、2、无穷大范数（np.inf）或其他正整数值。
            axis：           表示沿着哪个轴计算范数并进行归一化。
                                默认为-1。
    """
    n = np.linalg.norm(v, ord=ord, axis=axis, keepdims=True)
    n = np.where(n > 0, n, 1e-10)  # 防溢出
    return v / n


if __name__ == '__main__':
    print(normalize(np.array([[1, 2, 3, 4, 5]]), ord=2))



"""数据集生成器"""


def build_generator(executor_ls, size_ls=None):
    """
        根据 executor_ls 构建数据集的生成器

        参数：
            executor_ls：        执行器列表
            size_ls：            数据集大小
                                    当为非空时，将根据 size_ls 对 executor 实际上生成数据集的大小进行检查
        返回：
            generator：          一个产生 dataset 的生成器
    """
    assert isinstance(executor_ls, (list, tuple,))
    if size_ls is not None:
        assert isinstance(size_ls, (list, tuple,)) and len(size_ls) == len(executor_ls)

    def generator():
        for i, executor in enumerate(executor_ls):
            res = executor()
            if size_ls is None:
                assert len(res[list(res.keys())[0]]) == size_ls[i]
            yield res

    return generator()



import os
from code.utils.url import get_response
from code.spider.for_cvf import in_root_page, in_conf_page, in_home_page, in_item_page
from code.utils.file import my_abspath, read, write
import threading
import math


def parse_from_item_page(item_ls, i):
    """
        到对应的 item=item_ls[i] 中，访问 item["source_online"]["item_url"] 来获取相关信息，并补充至 item
    """
    item_url = item_ls[i]["source_online"]["item_url"]
    print(i, item_url)
    text = get_response(url=item_url)
    # 获取 abstract
    item_ls[i]["abstract"] = in_item_page.find_abstract(text)
    # 获取 paper_url
    item_ls[i]["source_online"]["paper_url"] = f'{root_url}/{in_item_page.find_paper(text)}'
    # 获取 authors
    item_ls[i]["authors"] = in_item_page.find_authors(text)


if __name__ == '__main__':
    # root
    root_url = "https://openaccess.thecvf.com"
    text = get_response(url=root_url)
    conference_ls = in_root_page.find_conferences(text)
    print(f"from root {root_url} \n"
          f"we get: conference_ls {conference_ls}")
    # conference_ls=conference_ls[7:]

    for conference in conference_ls:
        publisher = conference[:-4]
        year = conference[-4:]

        # conf
        conference_url = f"{root_url}/{conference}"
        text = get_response(url=conference_url)
        homepage_ls = in_conf_page.find_homepages(text, head=conference)
        if len(homepage_ls) == 0:
            homepage_ls = [conference]
        #
        print(f"from conference {publisher, year} \n"
              f"we get: homepage_ls {homepage_ls}")

        # home
        item_ls = []
        for homepage in homepage_ls:
            homepage_url = f"{root_url}/{homepage}"
            text = get_response(url=homepage_url)
            item_ls.extend(in_home_page.find_items(text))
        for item in item_ls:
            item["source_online"]["item_url"] = f'{root_url}/{item["source_online"]["item_url"]}'
        #
        temp = '\n'.join([f'{key}: {value}' for key, value in item_ls[0].items()])
        print(f"from homepage {homepage_ls} \n"
              f"we get: {len(item_ls)} items\n"
              f"for example one of them is:\n"
              f"{temp}")

        # item
        thread_nums = 400
        for i in range(math.ceil(len(item_ls) / thread_nums)):
            thread_ls = []
            for j in range(i * thread_nums, min((i + 1) * thread_nums, len(item_ls))):
                thread_ls.append(threading.Thread(target=parse_from_item_page, kwargs=dict(item_ls=item_ls, i=j)))
            for t in thread_ls:
                t.start()
            for t in thread_ls:
                t.join()

        # 补充 publisher year
        for item in item_ls:
            item["publisher"] = publisher
            item["year"] = year

        # 保存
        file_path = my_abspath(path=f"../../meta_data/{publisher}/{year}.json",
                               base=os.path.split(os.path.abspath(__file__))[0])
        os.makedirs(os.path.split(file_path)[0], exist_ok=True)
        write(path=file_path, content=item_ls, verbose=True)

        print(f"end --- {publisher} --- {root_url} --- {year}")



import numpy as np


def convert_format(**kwargs):
    """
        在各种格式的 坐标列表 之间进行转换

        参数：
            var:                输入的坐标列表。
            input_format:       <str> 描述输入的格式。
                                    要与 var 的真实格式对应。
                                    目前支持在以下三种坐标格式之间进行转换：
                                        "index_ls" , "indices_ls" , "zip_indices"
            output_format:      <str> 输出的目标格式。
            shape:              坐标所属的多维变量的形状。

        这三种坐标格式是什么？有什么区别？
            - index_ls      <np.array of integer> 坐标列表。
                                其中基本元素为 integer，每个 integer 对应着将多维变量打平 reshape(-1) 后某个元素的位置。
                                shape [n_num, ]
            - indices_ls    <np.array of np.array> 坐标列表。
                                其中每个 nparray 对应着将多维变量中某个元素的位置。
                                shape [n_num, index_num]
            - zip_indices   <tuple of np.array> 坐标列表。
                                是 indices_ls 进行维度变换后的产物。
                                shape [index_num, n_num]

            以 shape=(2,2) 的变量为例，则下面两种坐标表示方式指向的元素都是相同的：
                index_ls = [0, 1, 2, 3]
                indices_ls = [[0, 0], [0, 1], [1, 0], [1, 1]]
                zip_indices = ([0, 0, 1, 1], [0, 1, 0, 1])

        为什么要区分这三种坐标格式？
            这几种坐标有不同的使用场景，且这些场景还是比较高频出现的：
            - index_ls      在使用 np.transpose() 对 axis 进行转置，或者使用 dimension.transpose.inside_axis()
                                对 axis 内各个维度进行转置时，就需要使用 index_ls 形式的坐标来指定转置后各个 axis/维度 的位置。
            - indices_ls    遍历时有可能会用到。
            - zip_indices   适用于 numpy/torch 进行按照坐标的索引取值。
                                比如对于变量 x = [[0, 1], [2, 3]]，使用前面例子中的 zip_indices，进行下面形式的取值索引：
                                x[ zip_indices ] 就可以得到 [0, 1, 2, 3]
            如果还是看不明白，可以看一下本函数测试文件 test.py 来加深一下感受。
    """

    # 默认参数
    paras = {
        # 必要参数
        "var": None,
        "input_format": None,
        "output_format": None,
        "shape": None,
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    for key in {"input_format", "output_format"}:
        assert paras[key] in {"index_ls", "indices_ls", "zip_indices"}
    #
    if paras["input_format"] == "index_ls":
        assert paras["var"].ndim == 1
    elif paras["input_format"] == "indices_ls":
        assert paras["var"].ndim == 2
    else:
        assert isinstance(paras["var"], (tuple,))

    # 转换
    if paras["input_format"] == paras["output_format"]:
        res = paras["var"]
    elif paras["input_format"] == "index_ls":
        res = _index_to_indices(index_ls=paras["var"], shape=paras["shape"])
        if paras["output_format"] == "indices_ls":
            pass
        else:
            res = _indices_to_zip_type(indices_ls=res)
    elif paras["input_format"] == "indices_ls":
        if paras["output_format"] == "index_ls":
            res = _indices_to_index(indices_ls=paras["var"], shape=paras["shape"])
        else:
            res = _indices_to_zip_type(indices_ls=paras["var"])
    else:
        res = _zip_type_to_indices(zip_indices=paras["var"])
        if paras["output_format"] == "indices_ls":
            pass
        else:
            res = _indices_to_index(indices_ls=res, shape=paras["shape"])

    return res


def _indices_to_zip_type(indices_ls):
    # res = tuple([indices_ls[:, i].astype(dtype=np.int32) for i in range(indices_ls.shape[1])])
    res = tuple(np.moveaxis(indices_ls, 1, 0).astype(dtype=np.int32))
    return res


def _zip_type_to_indices(zip_indices):
    # res = np.stack(zip_indices, axis=1)
    res = np.moveaxis(zip_indices, 1, 0)
    return res


def _indices_to_index(indices_ls, shape):
    assert indices_ls.shape[1] == len(shape)

    source = np.arange(0, np.prod(shape)).astype(dtype=np.int32).reshape(shape)
    zip_indices = _indices_to_zip_type(indices_ls)
    index_ls = source[zip_indices].reshape(-1)
    return index_ls


def _index_to_indices(index_ls, shape):
    assert index_ls.size <= np.prod(shape)

    source = np.zeros(shape=shape)
    zip_indices = np.where(source >= 0)
    indices_ls = _zip_type_to_indices(zip_indices)
    indices_ls = indices_ls[index_ls]
    return indices_ls



from kevin_toolbox.math.dimension import coordinates
from kevin_toolbox.math import utils


def flatten_along_pattern(**kwargs):
    """
        将 x 的最后 dim_num 个维度按照 generate_func 指定的遍历顺序进行打平展开

        参数：
            x：                  <np.array/tensor>
            dim_num：            <integer>
            generate_func：      用于指定对 block 的遍历顺序
                                    默认使用 coordinates.generate(pattern="z_pattern", output_format="zip_indices") 进行遍历
                                    你也可以自定义一个根据参数 shape 生成 zip_indices 格式的坐标列表的函数，来指定遍历顺序
    """
    # 默认参数
    paras = {
        # 必要参数
        "x": None,
        "dim_num": None,
        "generate_func": lambda shape: coordinates.generate(pattern="z_pattern", output_format="zip_indices",
                                                            shape=shape),
    }

    # 获取参数
    paras.update(kwargs)

    # 校验参数
    _, function_table = utils.get_function_table_for_array_and_tensor(paras["x"])
    permute = function_table["permute"]
    x = paras["x"]
    assert isinstance(paras["dim_num"], (int,)) and 1 < paras["dim_num"] <= len(x.shape)
    dim_num = paras["dim_num"]
    # generate_func
    assert callable(paras["generate_func"])

    # 首先将 y 要被展开的维度提前
    # y: [64, 8, 4] ==> [8, 4, 64]
    dim_ls = list(range(len(x.shape)))
    dim_ls = dim_ls[-dim_num:] + dim_ls[:-dim_num]
    y = permute(x, dim_ls)

    # 按照指定 pattern 生成的 indices_ls 进行展开
    # y: [8, 4, 64] ==> [32, 64]
    indices = paras["generate_func"](shape=x.shape[-dim_num:])
    y = y[indices]

    # 恢复维度顺序
    # y: [32, 64] ==> [64, 32]
    dim_ls = list(range(len(y.shape)))
    dim_ls = dim_ls[1:] + [0]
    y = permute(y, dim_ls)

    return y



import numpy as np
from .parse_version import parse_version_string_to_array as parse_to_array


def sort_version_ls(version_ls, reverse=False, **kwargs):
    """
        对版本号组成的列表 version_ls 进行排序
            支持字符串形式或者数字列表/元组形式的版本号进行混合排序，
            比如： version_ls=["0.10.7", (0, 7), [0, 7, 5]]
            在排序后将得到： [(0, 7), [0, 7, 5], "0.10.7"]

        参数：
            version_ls:             <list of strings/arrays>
            reverse:                <boolean>
                                        默认为 False 表示从小到大排列，当设置为 True 时则反序。
        可选参数：
            mode：                   对齐方式
                                        "long"（默认）： 以 version_ls 中较长的版本号为基准进行对齐，对于较短的版本号，缺省部分将补 0
                                        "short"： 以较短的为基准进行对齐，对于较长的版本号，多余部分直接截断
                                        （建议选用 "long"，因为对于无法解释的 version，parse_to_array()
                                        默认返回 [0]，此时如果使用 "short" 可能引发意外的错误）
            sep：                    分隔符
                                        对于string类型的版本号，默认使用"."作为分隔符，但是也可以在 kwargs 中通过 sep 指定，
                                        具体参考 parse_to_array()
        返回：
            boolean 比较结果
    """
    assert isinstance(version_ls, (list, tuple,))
    mode = kwargs.get("mode", "long")
    assert mode in ["long", "short"]

    value_ls = []
    length_ls = []
    for version in version_ls:
        value = parse_to_array(version, **kwargs) if isinstance(version, (str,)) else version
        assert isinstance(value, (list, tuple, np.ndarray,))
        value_ls.append(list(value))
        length_ls.append(len(value))

    if mode == "long":
        target = max(length_ls)
        value_ls = [i + [0] * (target - l) for i, l in zip(value_ls, length_ls)]
    else:
        target = min(length_ls)
        value_ls = [i[:target] for i in value_ls]
    # sort
    pairs_ls = sorted(list(zip(value_ls, version_ls)), key=lambda x: x[0], reverse=reverse)
    #
    res = [i for _, i in pairs_ls]
    return res
